(*
 * Copyright 2005 Toshiyuki Maeda
 *
 * This file is part of TALK.
 *
 * TALK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * TALK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *)

open Types
open Integer_op

open Big_int
open Big_int_util

(*
 * Linear Integer Constraints Solver
 *)

type n_cop = N_EQ | N_NEQ | N_LE | N_GE

let reverse_n_cop = function
	| N_EQ ->
		N_EQ
	| N_NEQ ->
		N_NEQ
	| N_LE ->
		N_GE
	| N_GE ->
		N_LE
		
exception Apparently_unsatisfiable

type normalized = 
{
	n_array  : big_int array;
	n_cop : n_cop;
}


(*
 * Output functions for debugging
 *)
let output_n_ie ppf n_ie =
	let len = Array.length n_ie in
	for i = 0 to (len - 1) do
		let bi = n_ie.(i) in
		let s = string_of_big_int bi in
		if i = len - 1 then
			Format.fprintf ppf "%s" s
		else if bi <>^ bi_0 then
			Format.fprintf ppf "%s@ *@ X_%d@ +@ " s i
		else
			()
	done

let output_n_ic ppf n_ic =
	let ops = match n_ic.n_cop with
	| N_EQ ->
		"="
	| N_NEQ ->
		"<>"
	| N_LE ->
		"<="
	| N_GE ->
		">="
	in
	Format.fprintf ppf "@[%a@ %s@ 0@]@." output_n_ie n_ic.n_array ops

let output_n_ics ppf n_ics =
	List.iter (output_n_ic ppf) n_ics

let create_zero_n_ie num_tvs =
	Array.make (num_tvs + 1) bi_0

let create_n_ie n_ie =
	Array.copy n_ie


(*
 * Normalizers
 *)

let normalize_inte num_tvs tvs inte =
	let a = create_zero_n_ie num_tvs in

	(match inte with
	| IConst bi ->
		a.(num_tvs) <- bi
	| IVar tv ->
		let i = Util.find_index tv tvs in
		a.(i) <- bi_1
	)
	;
	a

let add_normalized_ie num_tvs n_ie1 n_ie2 =
	let n_ie = create_zero_n_ie num_tvs in
	for i = 0 to num_tvs do
		n_ie.(i) <- n_ie1.(i) +^ n_ie2.(i)
	done;
	n_ie

let rec is_constant_aux i num_tvs n_ie =
	if i >= num_tvs then
		true
	else

	let bi = n_ie.(i) in

	if bi <>^ bi_0 then
		false
	else
		is_constant_aux (i + 1) num_tvs n_ie

let is_constant num_tvs n_ie =
	is_constant_aux 0 num_tvs n_ie

let mul_normalized_ie_constant num_tvs n_ie bi =
	let n_ie_r = create_zero_n_ie num_tvs in
	for i = 0 to num_tvs do
		n_ie_r.(i) <- n_ie.(i) *^ bi
	done;
	n_ie_r

let mul_normalized_ie num_tvs n_ie1 n_ie2 =
	if is_constant num_tvs n_ie1 then
		mul_normalized_ie_constant num_tvs n_ie2 (n_ie1.(num_tvs))
	else if is_constant num_tvs n_ie2 then
		mul_normalized_ie_constant num_tvs n_ie1 (n_ie2.(num_tvs))
	else
		(* XXX : Non-linear constarint! *)
		raise Not_found

let merge_normalized_ie num_tvs aop n_ie1 n_ie2 =
	match aop with
	| A_ADD ->
		add_normalized_ie num_tvs n_ie1 n_ie2
	| A_MUL ->
		mul_normalized_ie num_tvs n_ie1 n_ie2

let rec normalize_ie num_tvs tvs = function
	| IEInte inte ->
		normalize_inte num_tvs tvs inte
	| IEArith (aop, ie1, ie2) ->
		let n_ie1 = normalize_ie num_tvs tvs ie1 in
		let n_ie2 = normalize_ie num_tvs tvs ie2 in
		merge_normalized_ie num_tvs aop n_ie1 n_ie2

let sub_normalized_ie num_tvs n_ie1 n_ie2 =
	let n_ie = create_zero_n_ie num_tvs in
	for i = 0 to num_tvs do
		n_ie.(i) <- n_ie1.(i) -^ n_ie2.(i)
	done;
	n_ie

let rec is_normalized_normalized_ie_aux i num_tvs n_ie =
	if i > num_tvs then
		(* == bi_0 *)
		true
	else

	let bi = n_ie.(i) in

	if bi =^ bi_0 then
		is_normalized_normalized_ie_aux (i + 1) num_tvs n_ie
	else
		bi >^ bi_0

let is_normalized_normalized_ie num_tvs n_ie =
	is_normalized_normalized_ie_aux 0 num_tvs n_ie

let negate_normalized_ie num_tvs n_ie =
	for i = 0 to num_tvs do
		n_ie.(i) <- (minus_big_int n_ie.(i))
	done

let get_non_zero_cos_aux i num_tvs n_ie =
	let lst = ref [] in
	for i = 0 to (num_tvs - 1) do
		let bi = n_ie.(i) in
		if bi <>^ bi_0 then
			lst := (bi :: !lst)
	done;
	!lst

let get_non_zero_cos num_tvs n_ie =
	get_non_zero_cos_aux 0 num_tvs n_ie

let rec get_gcd_bis = function
	| [] ->
		bi_1
	| [bi] ->
		abs_big_int bi
	| bi :: bis ->
		let gcd = get_gcd_bis bis in
		gcd_big_int bi gcd

let rec get_gcd num_tvs n_ie =
	let bis = get_non_zero_cos num_tvs n_ie in
	get_gcd_bis bis

let adjust_constant n_cop gcd num_tvs n_ie =
	let c = n_ie.(num_tvs) in
	let q, r = c /%^ gcd in

	match n_cop with
	| N_EQ ->
		if r <>^ bi_0 then
			raise Apparently_unsatisfiable
		else
			n_ie.(num_tvs) <- q
	| N_NEQ ->
		if r <>^ bi_0 then
			for i = 0 to num_tvs do
				n_ie.(i) <- bi_0
			done
		else
			n_ie.(num_tvs) <- q
	| N_LE ->
		n_ie.(num_tvs) <- div_ceil_big_int c gcd
	| N_GE ->
		n_ie.(num_tvs) <- div_floor_big_int c gcd

let normalize_normalized_ic num_tvs n_ic =
	let { n_array = n_ie; n_cop = n_cop; } = n_ic in

	let n_ie, n_cop =
		if is_normalized_normalized_ie num_tvs n_ie then
			n_ie, n_cop
		else
		begin
			negate_normalized_ie num_tvs n_ie;
			let n_cop = reverse_n_cop n_cop in
			n_ie, n_cop
		end
	in

	if is_constant num_tvs n_ie then
		{ n_array = n_ie; n_cop = n_cop; }
	else

	let gcd = get_gcd num_tvs n_ie in

	for i = 0 to (num_tvs - 1) do
		n_ie.(i) <- n_ie.(i) /^ gcd
	done;

	adjust_constant n_cop gcd num_tvs n_ie;

	{ n_array = n_ie; n_cop = n_cop; }

let normalize_normalized_ics num_tvs n_ics =
	List.map (normalize_normalized_ic num_tvs) n_ics

let create_normalized_ic num_tvs cop n_ie1 n_ie2 =

	(*
	 * signed/unsigned doesn't matter
	 * because signed-ness has been checked when type-checking Jcc instructions. 
	 *)
	let (cop, _) = cop in

	match cop with
	| C_EQ ->
		let n_ie = sub_normalized_ie num_tvs n_ie1 n_ie2 in
		{ n_array = n_ie; n_cop = N_EQ; }
	| C_NEQ ->
		let n_ie = sub_normalized_ie num_tvs n_ie1 n_ie2 in
		{ n_array = n_ie; n_cop = N_NEQ; }
	| C_LT ->
		let n_ie = sub_normalized_ie num_tvs n_ie1 n_ie2 in
		n_ie.(num_tvs) <- (n_ie.(num_tvs) +^ bi_1);
		{ n_array = n_ie; n_cop = N_LE; }
	| C_GT ->
		let n_ie = sub_normalized_ie num_tvs n_ie1 n_ie2 in
		n_ie.(num_tvs) <- (n_ie.(num_tvs) -^ bi_1);
		{ n_array = n_ie; n_cop = N_GE; }
	| C_LE ->
		let n_ie = sub_normalized_ie num_tvs n_ie1 n_ie2 in
		{ n_array = n_ie; n_cop = N_LE; }
	| C_GE ->
		let n_ie = sub_normalized_ie num_tvs n_ie1 n_ie2 in
		{ n_array = n_ie; n_cop = N_GE; }

let normalize_ic num_tvs tvs = function
	| ICCmp (cop, ie1, ie2) ->

		let n_ie1 = normalize_ie num_tvs tvs ie1 in
		let n_ie2 = normalize_ie num_tvs tvs ie2 in

		create_normalized_ic num_tvs cop n_ie1 n_ie2

let normalize_ics num_tvs tvs ics =
	List.fold_right (fun ic n_ics ->
	try
		let n_ic = normalize_ic num_tvs tvs ic in
		n_ic :: n_ics
	(* We discard non-linear constraints. *)
	with _ ->
		n_ics
	) ics []



(*
 * Obtain relations(lists) between TVs and Normalized TVs.
 *)
let get_tvs_list_inte tvs = function
	| IConst _ ->
		tvs
	| IVar tv ->
		if List.mem tv tvs then
			tvs
		else
			tvs @ [tv]

let rec get_tvs_list_ie tvs = function
	| IEInte inte ->
		get_tvs_list_inte tvs inte
	| IEArith (_, ie1, ie2) ->
		let tvs = get_tvs_list_ie tvs ie1 in
		get_tvs_list_ie tvs ie2

let get_tvs_list_ic tvs = function
	| ICCmp (_, ie1, ie2) ->
		let tvs = get_tvs_list_ie tvs ie1 in
		get_tvs_list_ie tvs ie2

let rec get_tvs_list tvs = function
	| [] ->
		tvs
	| ic :: ics ->
		let tvs = get_tvs_list_ic tvs ic in
		get_tvs_list tvs ics



(*
 * Routines for discarding unnecessary constraints.
 *)

let get_tvs_inte = function
	| IConst _ ->
		TVSet.empty
	| IVar tv ->
		TVSet.singleton tv

let rec get_tvs_ie = function
	| IEInte inte ->
		get_tvs_inte inte
	| IEArith (_, ie1, ie2) ->
		let tvs1 = get_tvs_ie ie1 in
		let tvs2 = get_tvs_ie ie2 in
		TVSet.union tvs1 tvs2

let get_tvs = function
	| ICCmp (_, ie1, ie2) ->
		let tvs1 = get_tvs_ie ie1 in
		let tvs2 = get_tvs_ie ie2 in
		TVSet.union tvs1 tvs2

let get_dependent_ics_one tvs ics =

	List.fold_right (fun ic (tvs, d_ics, i_ics) ->
		let tvs' = get_tvs ic in
		if TVSet.is_empty (TVSet.inter tvs tvs') then
			tvs, d_ics, ic :: i_ics
		else
			let tvs = TVSet.union tvs tvs' in
			tvs, (ic :: d_ics), i_ics
	) ics (tvs, [], [])

let rec get_dependent_ics_aux tvs ics =

	let tvs, d_ics, n_ics = get_dependent_ics_one tvs ics in

	if d_ics = [] then
		[]
	else
		let n_ics = get_dependent_ics_aux tvs n_ics in
		d_ics @ n_ics

let get_dependent_ics ic ics =

	let tvs = get_tvs ic in

	let ics = get_dependent_ics_aux tvs ics in

	ic :: ics


(*
 * Some utillity functions
 *)

let rec compare_n_ies_aux i num_tvs n_ie1 n_ie2 =
	if i > num_tvs then
		0
	else

	let c = compare_big_int n_ie1.(i) n_ie2.(i) in
	
	if c = 0 then
		compare_n_ies_aux (i + 1) num_tvs n_ie1 n_ie2
	else
		c	

let compare_n_ies num_tvs n_ie1 n_ie2 =
	compare_n_ies_aux 0 num_tvs n_ie1 n_ie2

let n_cop_to_int = function
	| N_EQ -> 0
	| N_NEQ -> 1
	| N_LE -> 2
	| N_GE -> 3

let compare_n_cop n_cop1 n_cop2 =
	let n1 = n_cop_to_int n_cop1 in
	let n2 = n_cop_to_int n_cop2 in
	n1 - n2

let compare_n_ics num_tvs n_ic1 n_ic2 =
	let { n_array = n_ie1; n_cop = n_cop1; } = n_ic1 in
	let { n_array = n_ie2; n_cop = n_cop2; } = n_ic2 in

	let c = compare_n_ies num_tvs n_ie1 n_ie2 in

	if c = 0 then
		compare_n_cop n_cop1 n_cop2
	else
		c

let sort_n_ics num_tvs n_ics =
	List.sort (compare_n_ics num_tvs) n_ics

let apparently_unsatisfiable num_tvs n_ic =
	let { n_array = n_ie; n_cop = n_cop; } = n_ic in

	if not (is_constant num_tvs n_ie) then
		false
	else

	let bi = n_ie.(num_tvs) in

	match n_cop with
	| N_EQ ->
		bi <>^ bi_0
	| N_NEQ ->
		bi =^ bi_0
	| N_LE ->
		bi >^ bi_0
	| N_GE ->
		bi <^ bi_0

	
let apparently_satisfiable num_tvs n_ic =
	let { n_array = n_ie; n_cop = n_cop; } = n_ic in

	if not (is_constant num_tvs n_ie) then
		false
	else

	let bi = n_ie.(num_tvs) in

	match n_cop with
	| N_EQ ->
		bi =^ bi_0
	| N_NEQ ->
		bi <>^ bi_0
	| N_LE ->
		bi <=^ bi_0
	| N_GE ->
		bi >=^ bi_0

let rec has_same_prefix_aux i num_tvs n_ie1 n_ie2 =

	if i >= num_tvs then
		true
	else
		let bi1 = n_ie1.(i) in
		let bi2 = n_ie2.(i) in

		if bi1 =^ bi2 then
			has_same_prefix_aux (i + 1) num_tvs n_ie1 n_ie2
		else
			false

let has_same_prefix num_tvs n_ic1 n_ic2 =
	has_same_prefix_aux 0 num_tvs n_ic1.n_array n_ic2.n_array

let rec n_ics_to_n_icss num_tvs = function
	| [] ->
		[]
	| n_ic :: n_ics ->
		let same_n_ics, n_ics = List.partition (has_same_prefix num_tvs n_ic) n_ics in
		let n_icss = n_ics_to_n_icss num_tvs n_ics in
		(n_ic :: same_n_ics) :: n_icss

let is_redundant_one (bi1, n_cop1) (bi2, n_cop2) =
	match (n_cop1, n_cop2) with
	| N_EQ, N_EQ | N_NEQ, N_NEQ->
		bi1 =^ bi2
	| N_NEQ, N_EQ ->
		bi1 <>^ bi2
	| N_NEQ, N_LE ->
		bi1 <^ bi2
	| N_NEQ, N_GE ->
		bi1 >^ bi2
	| N_LE, N_EQ | N_LE, N_LE ->
		bi1 <=^ bi2
	| N_GE, N_EQ | N_GE, N_GE ->
		bi1 >=^ bi2
	| _ ->
		false

let is_redundant num_tvs n_ic n_ics =
	let c = (n_ic.n_array.(num_tvs), n_ic.n_cop) in
	let cs = List.map (fun n_ic ->
		let { n_array = n_ie; n_cop = n_cop; } = n_ic in
		(n_ie.(num_tvs), n_cop)
	) n_ics in
	List.exists (is_redundant_one c) cs

let conflict (bi1, n_cop1) (bi2, n_cop2) =
	match (n_cop1, n_cop2) with
	| N_EQ, N_EQ ->
		bi1 <>^ bi2
	| N_EQ, N_NEQ | N_NEQ, N_EQ ->
		bi1 =^ bi2
	| N_EQ, N_LE | N_GE, N_EQ | N_GE, N_LE ->
		bi1 <^ bi2
	| N_EQ, N_GE | N_LE, N_EQ | N_LE, N_GE ->
		bi1 >^ bi2
	| _ ->
		false

let rec unsatisfiable_cs = function
	| [] ->
		false
	| c :: cs ->
		let b1 = List.exists (conflict c) cs in
		b1 || (unsatisfiable_cs cs)

let rec remove_redundant_n_ics num_tvs = function
	| [] ->
		[]
	| n_ic :: n_ics ->
		let n_ics = remove_redundant_n_ics num_tvs n_ics in
		if is_redundant num_tvs n_ic n_ics then
			n_ics
		else
			n_ic :: n_ics

let can_contract num_tvs n_ic1 n_ic2 =
	let n_cop1 = n_ic1.n_cop in
	let n_cop2 = n_ic2.n_cop in

	let bi1 = n_ic1.n_array.(num_tvs) in
	let bi2 = n_ic2.n_array.(num_tvs) in

	match n_cop1, n_cop2 with
	| N_LE, N_GE | N_GE, N_LE ->
		bi1 =^ bi2
	| _ ->
		false

let filter_and_add_eq num_tvs n_ic1 n_ics =
	let n_ics' = List.filter (fun n_ic2 ->
		not (can_contract num_tvs n_ic1 n_ic2)
	) n_ics in

	let len = List.length n_ics in
	let len' = List.length n_ics' in

	if len = len' then
		n_ic1 :: n_ics
	else (* some n_ic was contracted *)

	let n_ie = create_n_ie n_ic1.n_array in
	let n_ic = { n_array = n_ie; n_cop = N_EQ } in
	n_ic :: n_ics

let rec contract_n_ics num_tvs = function
	| [] ->
		[]
	| n_ic :: n_ics ->
		let n_ics = contract_n_ics num_tvs n_ics in
		filter_and_add_eq num_tvs n_ic n_ics

let apparently_unsatisfiable_n_ics num_tvs n_ics =
	let cs = List.map (fun n_ic ->
		let { n_array = n_ie; n_cop = n_cop; } = n_ic in
		(n_ie.(num_tvs), n_cop)
	) n_ics in
	unsatisfiable_cs cs

let cleanup_n_ics num_tvs n_ics =
	let n_ics = List.filter (fun n_ic ->
		not (apparently_satisfiable num_tvs n_ic)
	) n_ics in

	if n_ics = [] then
		[]
	else

	if List.exists (apparently_unsatisfiable num_tvs) n_ics then
		raise Apparently_unsatisfiable
	else

	let n_icss = n_ics_to_n_icss num_tvs n_ics in

	let n_icss = List.map (remove_redundant_n_ics num_tvs) n_icss in
	let n_icss = List.map (contract_n_ics num_tvs) n_icss in

	if n_icss = [] then
		[]
	else

	if List.exists (apparently_unsatisfiable_n_ics num_tvs) n_icss then
		raise Apparently_unsatisfiable
	else

	let n_ics = List.concat n_icss in

	n_ics

let is_eq num_tvs n_ic =
	n_ic.n_cop = N_EQ

let rec is_eq_ie_1_aux i num_tvs n_ie =
	if i >= num_tvs then
		false
	else
		if n_ie.(i) =^ bi_1 then
			true
		else
			is_eq_ie_1_aux (i + 1) num_tvs n_ie

let is_eq_ie_1 num_tvs n_ie =
	is_eq_ie_1_aux 0 num_tvs n_ie

let is_eq_1 num_tvs n_ic =
	n_ic.n_cop = N_EQ && is_eq_ie_1 num_tvs n_ic.n_array

let get_match matcher num_tvs n_ics =
	List.fold_right (fun n_ic (b, rst) ->
		if b = None then
			if matcher num_tvs n_ic then
				let n_ie = n_ic.n_array in
				(Some n_ie, rst)
			else
				(None, n_ic :: rst)
		else
			(b, n_ic :: rst)
	) n_ics (None, [])

let get_eq_any num_tvs n_ics =
	get_match is_eq num_tvs n_ics

let get_eq_1 num_tvs n_ics =
	get_match is_eq_1 num_tvs n_ics

let get_eq num_tvs n_ics =

	let b_opt, n_ics = get_eq_1 num_tvs n_ics in
	match b_opt with
	| None ->
		get_eq_any num_tvs n_ics
	| _ ->
		b_opt, n_ics

let get_coeffs num_tvs n_ie =
	let lst = ref [] in
	for i = 0 to (num_tvs - 1) do
		let a = n_ie.(i) in
		if a <>^ bi_0 then
			let a = abs_big_int a in
			lst := (i, a) :: !lst
	done;
	!lst

let get_smallest_coeff num_tvs n_ie =
	let coeffs = get_coeffs num_tvs n_ie in
	match coeffs with
	| [] ->
		failwith "A constant eq-constraint was not removed."
	| (i, m) :: lst ->
		let (k, _) = List.fold_left (fun (i, m) (j, n) ->
			if m >^ n then
				(j, n)
			else
				(i, m)
		) (i, m) lst in
		k

let special_mod a b =
	let bi_2 = bi 2 in
	let s = div_floor_big_int (bi_2 *^ a +^ b) (bi_2 *^ b) in
	a -^ (b *^ s)

let get_subst_src num_tvs n_ie j =
	let c = n_ie.(j) in
	let len = num_tvs + 1 in

	if c =^ bi_1 then
		Array.init len (fun i ->
			if i <> j then
				~^ (n_ie.(i))
			else
				bi_0
		)
	else if c =^ ~^ bi_1 then
		Array.init len (fun i ->
			if i <> j then
				n_ie.(i)
			else
				bi_0
		)
	else
		let sign = bi (sign_big_int c) in
		let m = (abs_big_int c) +^ bi_1 in
		Array.init len (fun i ->
			if i = j then
				~^ (sign *^ m)
			else
				let m' = special_mod n_ie.(i) m in
				sign *^ m'
		)

let substitute num_tvs j n_ie n_ic =
	let n_ie' = n_ic.n_array in

	let m = n_ie'.(j) in

	if m =^ bi_0 then
		()
	else

	for i = 0 to num_tvs do
		if i = j then
			n_ie'.(i) <- bi_0
		else
			n_ie'.(i) <- n_ie'.(i) +^ m *^ n_ie.(i)
	done

let remove_eq num_tvs n_ie n_ics =
	let j = get_smallest_coeff num_tvs n_ie in
	let n_ie' = get_subst_src num_tvs n_ie j in 
	List.iter (substitute num_tvs j n_ie') n_ics;
	n_ics

let get_tvs_n_ie num_tvs tvs n_ie =
	let r = ref tvs in

	for i = 0 to num_tvs do
		let bi = n_ie.(i) in
		if bi <>^ bi_0 && not (List.mem i tvs) then
			r := (i :: !r)
	done;

	!r

let get_tvs num_tvs n_ics =
	List.fold_left (fun tvs n_ic -> get_tvs_n_ie num_tvs tvs n_ic.n_array) [] n_ics

let get_num_of_tvs num_tvs n_ics =
	let tvs = get_tvs num_tvs n_ics in
	List.length tvs

let has_at_most_one_tv num_tvs n_ics =
	let n = get_num_of_tvs num_tvs n_ics in
	n <= 1

let convert_neq num_tvs n_ie =
	let n_ie1 = create_n_ie n_ie in
	let n_ie2 = create_n_ie n_ie in

	n_ie1.(num_tvs) <- n_ie1.(num_tvs) +^ bi_1;
	n_ie2.(num_tvs) <- n_ie1.(num_tvs) -^ bi_1;

	let n_ic1 = { n_array = n_ie1; n_cop = N_GE; } in
	let n_ic2 = { n_array = n_ie2; n_cop = N_LE; } in

	n_ic1, n_ic2

let rec remove_neq num_tvs = function
	| [] ->
		[[]]
	| n_ic :: n_ics ->
		let n_icss = remove_neq num_tvs n_ics in
		if n_ic.n_cop = N_NEQ then
			let n_ic1, n_ic2 = convert_neq num_tvs n_ic.n_array in
			let n_icss1 = List.map (fun n_ics -> n_ic1 :: n_ics) n_icss in
			let n_icss2 = List.map (fun n_ics -> n_ic2 :: n_ics) n_icss in
			n_icss1 @ n_icss2
		else
			List.map (fun n_ics -> n_ic :: n_ics) n_icss

let rec choose_real_tvs num_tvs n_ics =
	let r = ref [] in

	for i = 0 to (num_tvs - 1) do
		if List.exists (fun n_ic -> n_ic.n_array.(i) <>^ bi_0) n_ics then
			r := i :: (!r)
	done;
	!r

let upper_bounded_n_ic m n_ic =
	let bi = n_ic.n_array.(m) in

	if bi =^ bi_0 then
		false
	else

	if bi >^ bi_0 then
		let c = n_ic.n_cop in
		c = N_LE
	else
		let c = n_ic.n_cop in
		c = N_GE

let lower_bounded_n_ic m n_ic =
	let bi = n_ic.n_array.(m) in

	if bi =^ bi_0 then
		false
	else

	if bi <^ bi_0 then
		let c = n_ic.n_cop in
		c = N_LE
	else
		let c = n_ic.n_cop in
		c = N_GE

let upper_bounded n_ics tv =
	List.exists (upper_bounded_n_ic tv) n_ics

let lower_bounded n_ics tv =
	List.exists (lower_bounded_n_ic tv) n_ics

let rec choose_unbound_tv tvs n_ics =
	match tvs with
	| [] ->
		None
	| tv :: tvs ->
		if not (upper_bounded n_ics tv) || not (lower_bounded n_ics tv) then
			Some tv
		else
			choose_unbound_tv tvs n_ics

let is_exact_tv n_ics tv =
	List.for_all (fun n_ic ->
		let b = n_ic.n_array.(tv) in
		b =^ bi_0 || b =^ bi_1 || b =^ bi (-1)
	) n_ics

let rec choose_exact_tv tvs n_ics =
	match tvs with
	| [] ->
		None
	| tv :: tvs ->
		if is_exact_tv n_ics tv then
			Some tv
		else
			choose_unbound_tv tvs n_ics

let rec choose_any_tv tvs n_ics =
	let mi = List.fold_left (fun mi n_ic ->
		let n_ie = n_ic.n_array in
		List.fold_left (fun mi tv ->
			let b = n_ie.(tv) in
			match mi with
			| None ->
				Some (tv, b)
			| Some (tv', b') ->
				if b <^ b' then
					Some (tv, b)
				else
					mi
		) mi tvs
	) None n_ics in

	match mi with
	| None ->
		failwith "Something wrong in normalizers..."
	| Some (tv, _) ->
		tv

let choose_one_tv num_tvs n_ics =
	let tvs = choose_real_tvs num_tvs n_ics in

	(* choose an unbound variable *)
	let tv_opt = choose_unbound_tv tvs n_ics in

	match tv_opt with
	| Some tv ->
		0, tv
	| None ->

	(* choose an exact variable *)
	let tv_opt = choose_exact_tv tvs n_ics in

	match tv_opt with
	| Some tv ->
		1, tv
	| None ->

	(* choose a variable of the smallest coefficient *)
	let tv = choose_any_tv tvs n_ics in
		2, tv

let is_unbound_tv kind =
	kind = 0

let is_exact_tv kind =
	kind = 1

let remove_unbound_tv num_tvs n_ics tv =
	List.filter (fun n_ic ->
		let n_ie = n_ic.n_array in
		let b = n_ie.(tv) in
		b =^ bi_0
	) n_ics

let get_bounds_aux num_tvs tv (ls, us, os) n_ic =
	let n_ie = n_ic.n_array in
	let b = n_ie.(tv) in

	if b =^ bi_0 then
		(ls, us, n_ic :: os)
	else

	let cop = n_ic.n_cop in
	let n_ie = create_n_ie n_ie in
	n_ie.(tv) <- bi_0;

	if b >^ bi_0 then
	begin
		negate_normalized_ie num_tvs n_ie;
		if cop = N_GE then
			(b, n_ie) :: ls, us, os
		else (* must be N_LE *)
			ls, (b, n_ie) :: us, os
	end
	else (* b <^ bi_0 *)
		let b = minus_big_int b in
		if cop = N_LE then
			(b, n_ie) :: ls, us, os
		else (* must be N_GE *)
			ls, (b, n_ie) :: us, os

let get_bounds num_tvs n_ics tv =
	List.fold_left (get_bounds_aux num_tvs tv) ([], [], []) n_ics

let get_shadow combinator num_tvs ls us os =
	List.fold_left (fun rst lb ->
		List.fold_left (fun rst ub ->
			let n_ic = combinator num_tvs lb ub in
			n_ic :: rst
		) rst us
	) os ls

let combine_real_shadow num_tvs (a, n_ie_a) (b, n_ie_b) =
	let n_ie_a = mul_normalized_ie_constant num_tvs n_ie_a b in
	let n_ie_b = mul_normalized_ie_constant num_tvs n_ie_b a in
	let n_ie = sub_normalized_ie num_tvs n_ie_a n_ie_b in
	{ n_array = n_ie; n_cop = N_LE; }

let combine_dark_shadow num_tvs (a, n_ie_a) (b, n_ie_b) =
	let n_ie_a = mul_normalized_ie_constant num_tvs n_ie_a b in
	let n_ie_b = mul_normalized_ie_constant num_tvs n_ie_b a in
	let n_ie_c = create_zero_n_ie num_tvs in
	n_ie_c.(num_tvs) <- (a -^ bi_1) *^ (b -^ bi_1);
	let n_ie = sub_normalized_ie num_tvs n_ie_a n_ie_b in
	{ n_array = n_ie; n_cop = N_LE; }

let get_real_shadow = get_shadow combine_real_shadow
let get_dark_shadow = get_shadow combine_dark_shadow

let get_largest_upper_bound_aux mx us =
	List.fold_left (fun mx (a, _) ->
		if a >^ mx then
			a
		else
			mx
	) mx us
let get_largest_upper_bound = function
	| [] ->
		failwith "get_largest_upper_bound: there must be a bug in detecting unbounded variables."
	| (a, _) :: us ->
		get_largest_upper_bound_aux a us



(*
 * The solver
 *)

let rec solve num_tvs n_ics =
(*
	Format.printf "@[n_ics = %a@]@." output_n_ics n_ics;
*)
try
	let n_ics = cleanup_n_ics num_tvs n_ics in
	let b_opt, n_ics = get_eq num_tvs n_ics in

	match b_opt with
	| None ->
		(* no EQ *)
		if has_at_most_one_tv num_tvs n_ics then
			true
		else

		(* remove NEQ *)
		let n_icss = remove_neq num_tvs n_ics in
		List.exists (solve_inequalities num_tvs) n_icss
	| Some n_ie ->
		let n_ics = remove_eq num_tvs n_ie n_ics in
		let n_ics = normalize_normalized_ics num_tvs n_ics in
		solve num_tvs n_ics
with Apparently_unsatisfiable ->
	false

and solve_inequalities num_tvs n_ics =
	let kind, tv = choose_one_tv num_tvs n_ics in

	if is_unbound_tv kind then
		let n_ics = remove_unbound_tv num_tvs n_ics tv in
		solve num_tvs n_ics
	else


	let lower_bounds, upper_bounds, others = get_bounds num_tvs n_ics tv in

	if is_exact_tv kind then
		let real_shadow_n_ics = get_real_shadow num_tvs lower_bounds upper_bounds others in
		solve num_tvs real_shadow_n_ics
	else

	let real_shadow_n_ics = get_real_shadow num_tvs lower_bounds upper_bounds others in

	if not (solve num_tvs real_shadow_n_ics) then
		false
	else

	let dark_shadow_n_ics = get_dark_shadow num_tvs lower_bounds upper_bounds others in

	if solve num_tvs dark_shadow_n_ics then
		true
	else

	let a = get_largest_upper_bound upper_bounds in

	List.for_all (fun (b, n_ie_b) ->
		let n_ie' = create_n_ie n_ie_b in
		n_ie'.(tv) <- (bi_0 -^ b);
		let m = (a *^ b -^ a -^ b) /^ a in
		solve_inequalities_expensive num_tvs n_ie' n_ics m
	) lower_bounds

and solve_inequalities_expensive num_tvs n_ie n_ics m =
	if m <^ bi_0 then
		false
	else

	let n_ie' = create_n_ie n_ie in
	n_ie'.(num_tvs) <- n_ie'.(num_tvs) +^ m;
	let n_ic = { n_array = n_ie'; n_cop = N_EQ; } in
	let n_ics = n_ic :: n_ics in
	if solve num_tvs n_ics then
		true
	else
		solve_inequalities_expensive num_tvs n_ie n_ics (m -^ bi_1)

let is_satisfiable ics =
try
	let tvs = get_tvs_list [] ics in
	let num_tvs = List.length tvs in
	let num_ics = List.length ics in

	let n_ics = normalize_ics num_tvs tvs ics in
	let n_ics = normalize_normalized_ics num_tvs n_ics in
	let n_ics = sort_n_ics num_tvs n_ics in

	solve num_tvs n_ics

with Apparently_unsatisfiable ->
	false



(*
 * Entry points for the solver.
 *)

let implies ics1 ics2 =
(*
	Format.printf "@[?@ (%a)@ =>@ (%a)@]@."
		Output_types_raw.output_inte_cstrts ics1
		Output_types_raw.output_inte_cstrts ics2;
*)

	let b = List.exists (fun ic ->
		let ic = Integer_op.negate ic in
		let ics = get_dependent_ics ic ics1 in
		is_satisfiable ics
	) ics2 in

(*
	if b then
		Format.printf "@[?@ %a => %a@]@."
			Output_types_raw.output_inte_cstrts ics1
			Output_types_raw.output_inte_cstrts ics2;
*)

	not b

let implies_one ics1 ic =
	implies ics1 [ic]
