(*
 * Copyright 2005 Toshiyuki Maeda
 *
 * This file is part of TALK.
 *
 * TALK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * TALK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *)
 
open Types
open Annotations
open Pseudo_insts

(**************************************
	Convert routines
	from Human-Readable Type
	to De Bruijin's notation Type
 **************************************)

let rec find_tv_aux tvs name i =
	match tvs with
	| [] ->
		Util.failwithf "No such type variable %s." name
	| tv :: tvs ->
		let Readable_types.TV(name') = tv in
		if name = name'
		then
			i
		else 
			find_tv_aux tvs name (i + 1)

let find_tv tvs name = find_tv_aux tvs name 0

let convert_type_variable tvs tv =
	let Readable_types.TV(name) = tv in
	let i = find_tv tvs name in
	TV(i)

let rec convert_integer tvs inte =
	match inte with
(*
	| Readable_types.IConst (sign, str) ->
		let i = Nativeint.of_string str in
		let i = Big_int_util.big_int_of_nativeint_unsigned i in
		let i = if sign then
				Big_int.minus_big_int i
			else
				i
		in
		IConst i
*)
	| Readable_types.IConst str ->
		let sign, str = Util.chop_sign str in
		let i = Nativeint.of_string str in
		let i = Big_int_util.big_int_of_nativeint_unsigned i in
		let i = if sign then
				Big_int.minus_big_int i
			else
				i
		in
		IConst i
	| Readable_types.IVar tv ->
		let tv = convert_type_variable tvs tv in
		IVar tv

let convert_arith_op = function
	| Readable_types.A_ADD -> A_ADD
	| Readable_types.A_MUL -> A_MUL

let convert_cmp_op_raw = function
	| Readable_types.C_EQ -> C_EQ
	| Readable_types.C_NEQ -> C_NEQ
	| Readable_types.C_LT -> C_LT
	| Readable_types.C_GT -> C_GT
	| Readable_types.C_LE -> C_LE
	| Readable_types.C_GE -> C_GE

let convert_cmp_op cop = 
	let (cop_r, signed) = cop in
	let cop_r = convert_cmp_op_raw cop_r in
	(cop_r, signed)

let rec convert_integer_expr tvs inte_e =
	match inte_e with
	| Readable_types.IEInte inte ->
		let inte = convert_integer tvs inte in
		IEInte inte
	| Readable_types.IEArith (aop, inte_e1, inte_e2) ->
		let aop = convert_arith_op aop in
		let inte_e1 = convert_integer_expr tvs inte_e1 in
		let inte_e2 = convert_integer_expr tvs inte_e2 in
		IEArith (aop, inte_e1, inte_e2)

let add_integer_cstrt tvs ic ics =
	match ic with
	| Readable_types.ICCmp (cop, inte_e1, inte_e2) ->
		let cop = convert_cmp_op cop in
		let inte_e1 = convert_integer_expr tvs inte_e1 in
		let inte_e2 = convert_integer_expr tvs inte_e2 in
		let ic = ICCmp (cop, inte_e1, inte_e2) in
		Types_op.add_integer_cstrt ic ics

let convert_integer_cstrts tvs ics =
	List.fold_left (fun ics ic ->
		add_integer_cstrt tvs ic ics
	) Types_op.empty_integer_cstrts ics


let convert_kind kind =
	match kind with
	| Readable_types.KStack -> KStack
	| Readable_types.KInt -> KInt
	| Readable_types.KStore -> KStore

let convert_tv_to_kind (tv, kind) =
	let Readable_types.TV(name) = tv in
	let kind = convert_kind kind in
	(kind, name)

let convert_tvs_to_kinds c_tvs =
	List.rev_map convert_tv_to_kind c_tvs

let rec convert_stack_type tvs st =
	match st with
	| Readable_types.StkNull ->
		StkNull
	| Readable_types.StkVar tv ->
		let tv = convert_type_variable tvs tv in
		StkVar tv
	| Readable_types.StkCons (inte, st) ->
		let inte = convert_integer tvs inte in
		let st = convert_stack_type tvs st in
		StkCons (inte, st)

let convert_rich_stack_type tvs rstk =
	let stk = rstk.Readable_types.rstk_stk in
	let freelen = rstk.Readable_types.rstk_freelen in

	let stk = convert_stack_type tvs stk in
	let freelen = convert_integer tvs freelen in

	{ rstk_stk = stk; rstk_freelen = freelen; }

let convert_segment_size_type tvs seg_sz =
	match seg_sz with
	| Readable_types.SegSize16 ->
		SegSize16
	| Readable_types.SegSize32 ->
		SegSize32

let convert_segment_type tvs seg_typ =
	match seg_typ with
	| Readable_types.CodeSeg sz ->
		let sz = convert_segment_size_type tvs sz in
		CodeSeg sz
	| Readable_types.DataSeg sz ->
		let sz = convert_segment_size_type tvs sz in
		DataSeg sz

let convert_register_kind = function
	| Readable_types.RK_L -> RK_L
	| Readable_types.RK_H -> RK_H
	| Readable_types.RK_X -> RK_X
	| Readable_types.RK_E -> RK_E
	
let convert_register_type tvs reg_type =
	let (reg_kind, inte) = reg_type in
	let reg_kind = convert_register_kind reg_kind in
	let inte = convert_integer tvs inte in
	(reg_kind, inte)

let convert_flags_type tvs flags_type =
{
	f_rst = convert_integer tvs flags_type.Readable_types.f_rst;
	f_op1 = convert_integer tvs flags_type.Readable_types.f_op1;
	f_op2 = convert_integer tvs flags_type.Readable_types.f_op2;
	f_size = flags_type.Readable_types.f_size;
}

let convert_flags_type_opt tvs = function
	| None ->
		None
	| Some flags ->
		let flags = convert_flags_type tvs flags in
		Some flags

let convert_registers_type tvs regs_type =
{
	cs = convert_segment_type tvs regs_type.Readable_types.cs;

(*
	eax = convert_integer tvs reg_type.Readable_types.eax;
	ecx = convert_integer tvs reg_type.Readable_types.ecx;
	edx = convert_integer tvs reg_type.Readable_types.edx;
	ebx = convert_integer tvs reg_type.Readable_types.ebx;
	esi = convert_integer tvs reg_type.Readable_types.esi;
	edi = convert_integer tvs reg_type.Readable_types.edi;
	ebp = convert_integer tvs reg_type.Readable_types.ebp;
	esp = convert_integer tvs reg_type.Readable_types.esp;

	frt = convert_integer tvs reg_type.Readable_types.frt;
	fo1 = convert_integer tvs reg_type.Readable_types.fo1;
	fo2 = convert_integer tvs reg_type.Readable_types.fo2;
*)
	eax = convert_register_type tvs regs_type.Readable_types.eax;
	ecx = convert_register_type tvs regs_type.Readable_types.ecx;
	edx = convert_register_type tvs regs_type.Readable_types.edx;
	ebx = convert_register_type tvs regs_type.Readable_types.ebx;
	esi = convert_register_type tvs regs_type.Readable_types.esi;
	edi = convert_register_type tvs regs_type.Readable_types.edi;
	ebp = convert_register_type tvs regs_type.Readable_types.ebp;
	esp = convert_register_type tvs regs_type.Readable_types.esp;

	flags_opt = convert_flags_type_opt tvs regs_type.Readable_types.flags_opt;
}

let convert_tuple_type_elem tvs (sz, inte) =
	let inte = convert_integer tvs inte in
	(sz, inte)

let convert_tuple_type tvs tt =
	List.map (convert_tuple_type_elem tvs) tt

let rec convert_store tvs str =
	List.fold_left (fun str se ->
		add_store_elem tvs se str
	) Types_op.empty_store str

and convert_code_store tvs c_str =
	List.fold_left (fun c_str (inte, st) ->
		let inte = convert_integer tvs inte in
		let st = convert_code_type tvs st in
		IntMap.add inte st c_str
	) IntMap.empty c_str

and add_store_elem tvs elem str =
	match elem with
	| Readable_types.SEAtom atom ->
		let (inte, at) = convert_atom tvs atom in
		Types_op.add_store_location inte at str
	| Readable_types.SEVar tv ->
		let tv = convert_type_variable tvs tv in
		Types_op.add_store_variable tv str

and convert_atom tvs (inte, ht) =
	let inte = convert_integer tvs inte in
	let ht = convert_heap_type tvs ht in
	(inte, ht)

and convert_cstrt tvs cstrt =
	let ics = convert_integer_cstrts tvs cstrt.Readable_types.c_ics in
	let store = convert_store tvs cstrt.Readable_types.c_store in
	let code_store = convert_code_store tvs cstrt.Readable_types.c_code_store in

	let cstrt = { c_ics = ics; c_store = store; c_code_store = code_store; } in
	cstrt

and convert_cstrts tvs cstrts =
	List.map (convert_cstrt tvs) cstrts

and convert_prefix tvs prefix =
	let p_tvs = prefix.Readable_types.p_tvs in
	let (tvs', _) = List.split p_tvs in

	let tvs = List.rev_append tvs' tvs in
	let kinds = convert_tvs_to_kinds p_tvs in

	let cstrts = convert_cstrts tvs prefix.Readable_types.p_cstrts in

	let prefix = { p_kinds = kinds; p_cstrts = cstrts; } in
	(tvs, prefix)

and convert_state_type tvs st =
	let {
		Readable_types.st_prefix = prefix;
		Readable_types.st_registers = reg_typ;
	} = st in

	let (tvs, prefix) = convert_prefix tvs prefix in
	let reg_typ = convert_registers_type tvs reg_typ in

	{ st_prefix = prefix; st_registers = reg_typ; }

and convert_pack_type tvs pt =
	let {
		Readable_types.pt_prefix = prefix;
		Readable_types.pt_tuple = tt;
	} = pt in

	let (tvs, prefix) = convert_prefix tvs prefix in
	let tt = convert_tuple_type tvs tt in

	{ pt_prefix = prefix; pt_tuple = tt; }

and convert_concrete_type tvs ct =
	match ct with
	| Readable_types.Pack pt ->
		let pt = convert_pack_type tvs pt in
		Pack pt
	| Readable_types.Tuple tt ->
		let tt = convert_tuple_type tvs tt in
		Tuple tt

and convert_type tvs t =
	match t with
	| Readable_types.Name nt ->
		let nt = convert_named_type tvs nt in
		Name nt
	| Readable_types.Concrete ct ->
		let ct = convert_concrete_type tvs ct in
		Concrete ct

and convert_array_type tvs at =
	let t = at.Readable_types.at_type in
	let len = at.Readable_types.at_len in

	let t = convert_type tvs t in
	let len = convert_integer tvs len in

	{ at_type = t; at_len = len; }

and convert_heap_type tvs = function
	| Readable_types.Array at ->
		let at = convert_array_type tvs at in
		Array at
	| Readable_types.Stack rstk ->
		let rstk = convert_rich_stack_type tvs rstk in
		Stack rstk

and convert_code_type tvs ct =
	match ct with
	| Readable_types.NameCode nt ->
		let nt = convert_named_type tvs nt in
		NameCode nt
	| Readable_types.State st ->
		let st = convert_state_type tvs st in
		State st

and convert_named_type tvs nt =
	let (n, cs) = nt in
	let cs = List.map (convert_constructor tvs) cs in
	(n, cs)

and convert_constructor tvs = function
	| Readable_types.CStack st ->
		let st = convert_stack_type tvs st in
		CStack st
	| Readable_types.CInt inte ->
		let inte = convert_integer tvs inte in
		CInt inte
	| Readable_types.CStore str ->
		let str = convert_store tvs str in
		CStore str

let convert_named_type_body tvs = function
	| Readable_types.NtdConcrete ct ->
		let ct = convert_concrete_type tvs ct in
		NtdConcrete ct
	| Readable_types.NtdState st ->
		let st = convert_state_type tvs st in
		NtdState st

let convert_named_type_definition tvs ntd =
	let {
		Readable_types.ntd_tvs = n_tvs;
		Readable_types.ntd_body = bdy;
	} = ntd in

	let (tvs', _) = List.split n_tvs in

	let tvs = List.rev_append tvs' tvs in
	let kinds = convert_tvs_to_kinds n_tvs in

	let bdy = convert_named_type_body tvs bdy in

	{ ntd_kinds = kinds; ntd_body = bdy; }

let convert_ad tvs ad =
	let base = convert_integer tvs ad.Readable_annotations.ad_base in
(*
	let scale = ad.Readable_annotations.ad_scale in
	let index = convert_integer tvs ad.Readable_annotations.ad_index in
*)
	let offset = ad.Readable_annotations.ad_offset in

(*
	{ ad_base = base; ad_scale = scale; ad_index = index; ad_offset = offset; }
*)
	{ ad_base = base; ad_offset = offset; }

let convert_p_op tvs = function
	| Readable_pseudo_insts.P_Op_Mem (m_op, ad) ->
		let ad = convert_ad tvs ad in
		P_Op_Mem (m_op, ad)
	| Readable_pseudo_insts.P_Op_Imd (bi, sz) ->
		P_Op_Imd (bi, sz)
	| Readable_pseudo_insts.P_Op_Reg reg ->
		P_Op_Reg reg

let convert_pseudo_inst tvs = function
	| Readable_pseudo_insts.P_ROLL (inte, nt) ->
		let inte = convert_integer tvs inte in
		let nt = convert_named_type tvs nt in
		P_ROLL (inte, nt)
	| Readable_pseudo_insts.P_UNROLL inte ->
		let inte = convert_integer tvs inte in
		P_UNROLL inte
	| Readable_pseudo_insts.P_PACK (inte, pt, cs) ->
		let inte = convert_integer tvs inte in
		let pt = convert_pack_type tvs pt in
		let cs = List.map (convert_constructor tvs) cs in
		P_PACK (inte, pt, cs)
	| Readable_pseudo_insts.P_UNPACK (dbg_info, inte) ->
		let inte = convert_integer tvs inte in
		P_UNPACK (dbg_info, inte)
	| Readable_pseudo_insts.P_FALL ->
		P_FALL
	| Readable_pseudo_insts.P_SPLIT (dbg_info, inte1, inte2) ->
		let inte1 = convert_integer tvs inte1 in
		let inte2 = convert_integer tvs inte2 in
		P_SPLIT (dbg_info, inte1, inte2)
	| Readable_pseudo_insts.P_CONCAT (dbg_info, inte1, inte2) ->
		let inte1 = convert_integer tvs inte1 in
		let inte2 = convert_integer tvs inte2 in
		P_CONCAT (dbg_info, inte1, inte2)
	| Readable_pseudo_insts.P_TUPLE_SPLIT (dbg_info, inte1, inte2) ->
		let inte1 = convert_integer tvs inte1 in
		let inte2 = convert_integer tvs inte2 in
		P_TUPLE_SPLIT (dbg_info, inte1, inte2)
	| Readable_pseudo_insts.P_TUPLE_CONCAT (inte1, inte2) ->
		let inte1 = convert_integer tvs inte1 in
		let inte2 = convert_integer tvs inte2 in
		P_TUPLE_CONCAT (inte1, inte2)
	| Readable_pseudo_insts.P_CALLN p_op ->
		let p_op = convert_p_op tvs p_op in
		P_CALLN p_op
	| Readable_pseudo_insts.P_JMPN p_op ->
		let p_op = convert_p_op tvs p_op in
		P_JMPN p_op
	| Readable_pseudo_insts.P_RET ->
		P_RET
	| Readable_pseudo_insts.P_PUSH (dbg_info, p_op) ->
		let p_op = convert_p_op tvs p_op in
		P_PUSH (dbg_info, p_op)
	| Readable_pseudo_insts.P_PUSHF dbg_info ->
		P_PUSHF dbg_info
	| Readable_pseudo_insts.P_MOV (p_op1, p_op2) ->
		let p_op1 = convert_p_op tvs p_op1 in
		let p_op2 = convert_p_op tvs p_op2 in
		P_MOV (p_op1, p_op2)
	| Readable_pseudo_insts.P_ADD (dbg_info, p_op1, p_op2) ->
		let p_op1 = convert_p_op tvs p_op1 in
		let p_op2 = convert_p_op tvs p_op2 in
		P_ADD (dbg_info, p_op1, p_op2)
	| Readable_pseudo_insts.P_SUB (dbg_info, p_op1, p_op2) ->
		let p_op1 = convert_p_op tvs p_op1 in
		let p_op2 = convert_p_op tvs p_op2 in
		P_SUB (dbg_info, p_op1, p_op2)
	| Readable_pseudo_insts.P_CMP (dbg_info, p_op1, p_op2) ->
		let p_op1 = convert_p_op tvs p_op1 in
		let p_op2 = convert_p_op tvs p_op2 in
		P_CMP (dbg_info, p_op1, p_op2)
	| Readable_pseudo_insts.P_DEC (dbg_info, p_op) ->
		let p_op = convert_p_op tvs p_op in
		P_DEC (dbg_info, p_op)
	| Readable_pseudo_insts.P_IMUL (dbg_info, p_op1, p_op2, p_op3) ->
		let p_op1 = convert_p_op tvs p_op1 in
		let p_op2 = convert_p_op tvs p_op2 in
		let p_op3 = convert_p_op tvs p_op3 in
		P_IMUL (dbg_info, p_op1, p_op2, p_op3)
	| Readable_pseudo_insts.P_POP (dbg_info, p_op) ->
		let p_op = convert_p_op tvs p_op in
		P_POP (dbg_info, p_op)
	| Readable_pseudo_insts.P_POPF dbg_info ->
		P_POPF dbg_info
	| Readable_pseudo_insts.P_Jcc (cc, p_op) ->
		let p_op = convert_p_op tvs p_op in
		P_Jcc (cc, p_op)

	| Readable_pseudo_insts.P_CAST_TO_TUPLE (dbg_info, inte) ->
		let inte = convert_integer tvs inte in
		P_CAST_TO_TUPLE (dbg_info, inte)
	| Readable_pseudo_insts.P_CAST_TO_ARRAY (dbg_info, inte) ->
		let inte = convert_integer tvs inte in
		P_CAST_TO_ARRAY (dbg_info, inte)
	| Readable_pseudo_insts.P_CAST_TO_STACK (dbg_info, inte) ->
		let inte = convert_integer tvs inte in
		P_CAST_TO_STACK (dbg_info, inte)
