(*
 * Copyright 2005 Toshiyuki Maeda
 *
 * This file is part of TALK.
 *
 * TALK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * TALK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *)

open Instructions
open Instructions_util
open Disassemble_common
open Util

let get_disas_gen_reg_only inst i =
	fun dctx seg off ->
		let opsize = match dctx.dctx_op_attr with
			| OpSizeAttr16 -> OpSize16
			| OpSizeAttr32 -> OpSize32 in
		off, inst (get_genreg_op opsize i)

let get_disas_INC index = get_disas_gen_reg_only inst_INC index
let get_disas_DEC index = get_disas_gen_reg_only inst_DEC index
let get_disas_PUSH index = get_disas_gen_reg_only inst_PUSH index
let get_disas_POP index = get_disas_gen_reg_only inst_POP index

let get_operand dctx top seg off =
	match top with
	| (AM_O, ot) -> get_operand_O dctx ot seg off
	| (AM_I, ot) -> get_operand_I dctx ot seg off
	| (AM_J, ot) -> get_operand_J dctx ot seg off
        | (a, _) -> failwithf "get_operand: (%a) has not been supported yet." output_admethod a

let get_disas_Jcc index top dctx seg off =
	let (off, op) = get_operand dctx top seg off in
	let cc = get_cc index in
	off, inst_Jcc cc op

let get_disas_PUSH_imd top dctx seg off =
	let (off, op) = get_operand dctx top seg off in
	off, inst_PUSH op

let get_disas_XXX_gen inst reg top dctx seg off =
	let op1 = Op_Reg reg in
	let (off, op2) = get_operand dctx top seg off in
	off, inst op1 op2

let get_disas_XXX_eXX disas_gen index top dctx seg off =
	let op_size = ot_to_op_size OT_v dctx.dctx_op_attr in
	let reg = get_genreg op_size index in
	disas_gen reg top dctx seg off

let get_disas_ADD_gen = get_disas_XXX_gen inst_ADD
let get_disas_ADD_eXX = get_disas_XXX_eXX get_disas_ADD_gen

let get_disas_ADD_AL = get_disas_ADD_gen (Reg8 AL)
let get_disas_ADD_eAX = get_disas_ADD_eXX 0

let get_disas_CMP_gen = get_disas_XXX_gen inst_CMP
let get_disas_CMP_eXX = get_disas_XXX_eXX get_disas_CMP_gen

let get_disas_CMP_AL = get_disas_CMP_gen (Reg8 AL)
let get_disas_CMP_eAX = get_disas_CMP_eXX 0

let get_disas_PUSHF dctx seg off = off, Inst_PUSHF
let get_disas_POPF dctx seg off = off, Inst_POPF
let get_disas_RET dctx seg off = off, Inst_RET

let get_disas_MOV_to_gen = get_disas_XXX_gen inst_MOV
let get_disas_MOV_to_eXX = get_disas_XXX_eXX get_disas_MOV_to_gen

let get_disas_MOV_from_gen reg top dctx seg off =
	let op2 = Op_Reg reg in
	let (off, op1) = get_operand dctx top seg off in
	off, inst_MOV op1 op2

let get_disas_MOV_from_eXX index top dctx seg off =
	let op_size = ot_to_op_size OT_v dctx.dctx_op_attr in
	let reg = get_genreg op_size index in
	get_disas_MOV_from_gen reg top dctx seg off

let get_disas_MOV_to_AL = get_disas_MOV_to_gen (Reg8 AL)
let get_disas_MOV_to_eAX = get_disas_MOV_to_eXX 0
let get_disas_MOV_to_eCX = get_disas_MOV_to_eXX 1
let get_disas_MOV_to_eDX = get_disas_MOV_to_eXX 2
let get_disas_MOV_to_eBX = get_disas_MOV_to_eXX 3
let get_disas_MOV_to_eSP = get_disas_MOV_to_eXX 4
let get_disas_MOV_to_eBP = get_disas_MOV_to_eXX 5
let get_disas_MOV_to_eSI = get_disas_MOV_to_eXX 6
let get_disas_MOV_to_eDI = get_disas_MOV_to_eXX 7

let get_disas_MOV_from_AL = get_disas_MOV_from_gen (Reg8 AL)
let get_disas_MOV_from_eAX go = get_disas_MOV_from_eXX 0 go
let get_disas_MOV_from_eCX go = get_disas_MOV_from_eXX 1 go
let get_disas_MOV_from_eDX go = get_disas_MOV_from_eXX 2 go
let get_disas_MOV_from_eBX go = get_disas_MOV_from_eXX 3 go
let get_disas_MOV_from_eSP go = get_disas_MOV_from_eXX 4 go
let get_disas_MOV_from_eBP go = get_disas_MOV_from_eXX 5 go
let get_disas_MOV_from_eSI go = get_disas_MOV_from_eXX 6 go
let get_disas_MOV_from_eDI go = get_disas_MOV_from_eXX 7 go

let get_disas_MOV_to_CL = get_disas_MOV_to_gen (Reg8 CL)
let get_disas_MOV_to_DL = get_disas_MOV_to_gen (Reg8 DL)
let get_disas_MOV_to_BL = get_disas_MOV_to_gen (Reg8 BL)
let get_disas_MOV_to_AH = get_disas_MOV_to_gen (Reg8 AH)
let get_disas_MOV_to_CH = get_disas_MOV_to_gen (Reg8 CH)
let get_disas_MOV_to_DH = get_disas_MOV_to_gen (Reg8 DH)
let get_disas_MOV_to_BH = get_disas_MOV_to_gen (Reg8 BH)

let get_disas_CALLN dctx seg off =
	let (off, op) = get_operand_J dctx OT_v seg off in
	off, inst_CALLN op

let get_disas_JMPN ot dctx seg off =
	let (off, op) = get_operand_J dctx ot seg off in
	off, inst_JMPN op

let get_operand_Ap dctx seg off =
	let (off, op2) = get_operand_I dctx OT_v seg off in
	let (off, op1) = get_operand_I dctx OT_w seg off in
	off, op1, op2

let get_disas_JMPFI dctx seg off =
	let (off, op1, op2) = get_operand_Ap dctx seg off in
	off, inst_JMPFI op1 op2

let get_1byte_nomodrm_op_raw b =
	match b with
	| 0x04 -> get_disas_ADD_AL oi_Ib
	| 0x05 -> get_disas_ADD_eAX oi_Iv

	| 0x3C -> get_disas_CMP_AL oi_Ib
	| 0x3D -> get_disas_CMP_eAX oi_Iv

	| 0x40 | 0x41 | 0x42 | 0x43 | 0x44 | 0x45 | 0x46 | 0x47 -> get_disas_INC (b - 0x40)
	| 0x48 | 0x49 | 0x4A | 0x4B | 0x4C | 0x4D | 0x4E | 0x4F -> get_disas_DEC (b - 0x48)

	| 0x50 | 0x51 | 0x52 | 0x53 | 0x54 | 0x55 | 0x56 | 0x57 -> get_disas_PUSH (b - 0x50)
	| 0x58 | 0x59 | 0x5A | 0x5B | 0x5C | 0x5D | 0x5E | 0x5F -> get_disas_POP (b - 0x58)

	| 0x68 -> get_disas_PUSH_imd oi_Iv
	| 0x6A -> get_disas_PUSH_imd oi_Ib

	| 0x70 | 0x71 | 0x72 | 0x73 | 0x74 | 0x75 | 0x76 | 0x77
	| 0x78 | 0x79 | 0x7A | 0x7B | 0x7C | 0x7D | 0x7E | 0x7F -> get_disas_Jcc (b - 0x70) oi_Jb

	| 0x9C -> get_disas_PUSHF
	| 0x9D -> get_disas_POPF

	| 0xA0 -> get_disas_MOV_to_AL oi_Ob
	| 0xA1 -> get_disas_MOV_to_eAX oi_Ov
	| 0xA2 -> get_disas_MOV_from_AL oi_Ob
	| 0xA3 -> get_disas_MOV_from_eAX oi_Ov

	| 0xB0 -> get_disas_MOV_to_AL oi_Ib
	| 0xB1 -> get_disas_MOV_to_CL oi_Ib
	| 0xB2 -> get_disas_MOV_to_DL oi_Ib
	| 0xB3 -> get_disas_MOV_to_BL oi_Ib
	| 0xB4 -> get_disas_MOV_to_AH oi_Ib
	| 0xB5 -> get_disas_MOV_to_CH oi_Ib
	| 0xB6 -> get_disas_MOV_to_DH oi_Ib
	| 0xB7 -> get_disas_MOV_to_BH oi_Ib

	| 0xB8 -> get_disas_MOV_to_eAX oi_Iv
	| 0xB9 -> get_disas_MOV_to_eCX oi_Iv
	| 0xBA -> get_disas_MOV_to_eDX oi_Iv
	| 0xBB -> get_disas_MOV_to_eBX oi_Iv
	| 0xBC -> get_disas_MOV_to_eSP oi_Iv
	| 0xBD -> get_disas_MOV_to_eBP oi_Iv
	| 0xBE -> get_disas_MOV_to_eSI oi_Iv
	| 0xBF -> get_disas_MOV_to_eDI oi_Iv

	| 0xC3 -> get_disas_RET

	| 0xE8 -> get_disas_CALLN

	| 0xE9 -> get_disas_JMPN OT_v
	| 0xEA -> get_disas_JMPFI
	| 0xEB -> get_disas_JMPN OT_b

	| _ -> reserved_inst "Unknown opcode = 0x%02x" b

let get_1byte_nomodrm_op_0F_raw b =
	match b with
	| 0x80 | 0x81 | 0x82 | 0x83 | 0x84 | 0x85 | 0x86 | 0x87
	| 0x88 | 0x89 | 0x8A | 0x8B | 0x8C | 0x8D | 0x8E | 0x8F -> get_disas_Jcc (b - 0x80) oi_Jv
	| _ -> reserved_inst "Unknown opcode = 0x0f%02x" b

let get_1byte_no_modrm_op b = get_1byte_nomodrm_op_raw b

let get_2byte_no_modrm_op_0F b = get_1byte_nomodrm_op_0F_raw b
