(*
 * Copyright 2005 Toshiyuki Maeda
 *
 * This file is part of TALK.
 *
 * TALK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * TALK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *)

(* Machine-dependent (for IA32) *)

open Nativeint
open Machine
open Elf_common

open Util

type elf_header =
{
	e_ident		: elf_ident;
	e_type		: elf_type;
	e_machine	: elf_machine;
	e_version	: elf_version;
	e_entry		: elf32_addr;
	e_phoff		: elf32_off;
	e_shoff		: elf32_off;
	e_flags		: elf32_word;
	e_ehsize	: elf32_half;
	e_phentsize	: elf32_half;
	e_phnum		: elf32_half;
	e_shentsize	: elf32_half;
	e_shnum		: elf32_half;
	e_shstrndx	: elf32_half;
}

and elf_ident =
{
	ei_class	: elf_class;
	ei_data		: elf_data;
	ei_version	: elf_version;
}

and elf_type =
	| ET_REL
	| ET_EXEC
	| ET_DYN
	| ET_CORE

and elf_machine =
	| EM_M32
	| EM_SPARC
	| EM_386
	| EM_68K
	| EM_88K
	| EM_860
	| EM_MIPS
	| EM_MIPS_RS4_BE

and elf_version =
	| EV_CURRENT

and elf_class =
	| ELFCLASS32
	| ELFCLASS64

and elf_data =
	| ELFDATA2LSB
	| ELFDATA2MSB

let fi_ident	= { fi_off = 0;	fi_size = 16; }
let fi_type	= { fi_off = next_off fi_ident;		fi_size = elf32_half_size; }
let fi_machine	= { fi_off = next_off fi_type;		fi_size = elf32_half_size; }
let fi_version	= { fi_off = next_off fi_machine;	fi_size = elf32_word_size; }
let fi_entry	= { fi_off = next_off fi_version;	fi_size = elf32_addr_size; }
let fi_phoff	= { fi_off = next_off fi_entry;		fi_size = elf32_off_size; }
let fi_shoff	= { fi_off = next_off fi_phoff;		fi_size = elf32_off_size; }
let fi_flags	= { fi_off = next_off fi_shoff;		fi_size = elf32_word_size; }
let fi_ehsize	= { fi_off = next_off fi_flags;		fi_size = elf32_half_size; }
let fi_phentsize= { fi_off = next_off fi_ehsize;	fi_size = elf32_half_size; }
let fi_phnum	= { fi_off = next_off fi_phentsize;	fi_size = elf32_half_size; }
let fi_shentsize= { fi_off = next_off fi_phnum;		fi_size = elf32_half_size; }
let fi_shnum	= { fi_off = next_off fi_shentsize;	fi_size = elf32_half_size; }
let fi_shstrndx	= { fi_off = next_off fi_shnum;		fi_size = elf32_half_size; }

let fi_ident_class	= { fi_off = fi_ident.fi_off + 4; fi_size = 1; }
let fi_ident_data	= { fi_off = fi_ident.fi_off + 5; fi_size = 1; }
let fi_ident_version	= { fi_off = fi_ident.fi_off + 6; fi_size = 1; }

let get_ident_class seg =
	let v = get_field seg fi_ident_class in
	let i = to_int v in
	match i with
	| 1 -> ELFCLASS32
	| 2 -> ELFCLASS64
	| _ -> failwithf "invalid ELF_ident.class = %d." i

let get_ident_data seg =
	let v = get_field seg fi_ident_data in
	let i = to_int v in
	match i with
	| 1 -> ELFDATA2LSB
	| 2 -> ELFDATA2MSB
	| _ -> failwithf "invalid ELF ident.data = %d." i

let get_ident_version seg =
	let v = get_field seg fi_ident_version in
	let i = to_int v in
	match i with
	| 1 -> EV_CURRENT
	| _ -> failwithf "invalid ELF ident.version = %d." i

let get_ident seg =
	let ei_class = get_ident_class seg in
	let ei_data = get_ident_data seg in
	let ei_version = get_ident_version seg in
	{
		ei_class = ei_class;
		ei_data = ei_data;
		ei_version = ei_version;
	}


let get_type seg =
	let v = get_field seg fi_type in
	let i = to_int v in
	match i with
	| 1 -> ET_REL
	| 2 -> ET_EXEC
	| 3 -> ET_DYN
	| 4 -> ET_CORE
	| _ -> failwithf "invalid ELF type = %d." i

let get_machine seg =
	let v = get_field seg fi_machine in
	let i = to_int v in
	match i with
	| 1 -> EM_M32
	| 2 -> EM_SPARC
	| 3 -> EM_386
	| 4 -> EM_68K
	| 5 -> EM_88K
	| 7 -> EM_860
	| 8 -> EM_MIPS
	| 10 -> EM_MIPS_RS4_BE
	| _ -> failwithf "invalid ELF machine = %d." i

let get_version seg =
	let v = get_field seg fi_version in
	let i = to_int v in
	match i with
	| 1 -> EV_CURRENT
	| _ -> failwithf "invalid ELF version = %d." i

let get_entry seg = get_field seg fi_entry
let get_phoff seg = get_field seg fi_phoff
let get_shoff seg = get_field seg fi_shoff
let get_flags seg = get_field seg fi_flags
let get_ehsize seg = get_field seg fi_ehsize
let get_phentsize seg = get_field seg fi_phentsize
let get_phnum seg = get_field seg fi_phnum
let get_shentsize seg = get_field seg fi_shentsize
let get_shnum seg = get_field seg fi_shnum
let get_shstrndx seg = get_field seg fi_shstrndx


let get_header seg =
	let e_ident = get_ident seg in
	let e_type = get_type seg in
	let e_machine = get_machine seg in
	let e_version = get_version seg in
	let e_entry = get_entry seg in
	let e_phoff = get_phoff seg in
	let e_shoff = get_shoff seg in
	let e_flags = get_flags seg in
	let e_ehsize = get_ehsize seg in
	let e_phentsize = get_phentsize seg in
	let e_phnum = get_phnum seg in
	let e_shentsize = get_shentsize seg in
	let e_shnum = get_shnum seg in
	let e_shstrndx = get_shstrndx seg in
	{
		e_ident = e_ident;
		e_type = e_type;
		e_machine = e_machine;
		e_version = e_version;
		e_entry = e_entry;
		e_phoff = e_phoff;
		e_shoff = e_shoff;
		e_flags = e_flags;
		e_ehsize = e_ehsize;
		e_phentsize = e_phentsize;
		e_phnum = e_phnum;
		e_shentsize = e_shentsize;
		e_shnum = e_shnum;
		e_shstrndx = e_shstrndx;
	}

