(*
 * Copyright 2005 Toshiyuki Maeda
 *
 * This file is part of TALK.
 *
 * TALK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * TALK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *)

open Nativeint
open Machine
open Elf_common
open Elf_header

open Util

type program_header =
{
	p_type		: segment_type;
	p_offset	: elf32_off;
	p_vaddr		: elf32_addr;
	p_paddr		: elf32_addr;
	p_filesz	: elf32_word;
	p_memsz		: elf32_word;
	p_flags		: segment_flag;
	p_align		: elf32_word;
}

and segment_type =
	| PT_LOAD
	| PT_DYNAMIC
	| PT_INTERP
	| PT_NOTE
	| PT_SHLIB
	| PT_PHDR
	| PT_GNU_EH_FRAME
	| PT_STACK

and segment_flag =
{
	phf_x : bool;
	phf_w : bool;
	phf_r : bool;
}

let fi_type	= { fi_off = 0;				fi_size = elf32_word_size; }
let fi_offset	= { fi_off = next_off fi_type;		fi_size = elf32_off_size; }
let fi_vaddr	= { fi_off = next_off fi_offset;	fi_size = elf32_addr_size; }
let fi_paddr	= { fi_off = next_off fi_vaddr;		fi_size = elf32_addr_size; }
let fi_filesz	= { fi_off = next_off fi_paddr;		fi_size = elf32_word_size; }
let fi_memsz	= { fi_off = next_off fi_filesz;	fi_size = elf32_word_size; }
let fi_flags	= { fi_off = next_off fi_memsz;		fi_size = elf32_word_size; }
let fi_align	= { fi_off = next_off fi_flags;		fi_size = elf32_word_size; }

let phf_x_bit = 0x1
let phf_w_bit = 0x2
let phf_r_bit = 0x4

let get_type seg =
	let v = get_field seg fi_type in
	let i = to_int v in
	match i with
	| 1 -> PT_LOAD
	| 2 -> PT_DYNAMIC
	| 3 -> PT_INTERP
	| 4 -> PT_NOTE
	| 5 -> PT_SHLIB
	| 6 -> PT_PHDR
	| 0x6474e550 -> PT_GNU_EH_FRAME
	| 0x6474e551 -> PT_STACK
	| _ -> failwithf "invalid ELF program header type = %d." i

let get_offset seg = get_field seg fi_offset
let get_vaddr seg = get_field seg fi_vaddr
let get_paddr seg = get_field seg fi_paddr
let get_filesz seg = get_field seg fi_filesz
let get_memsz seg = get_field seg fi_memsz

let get_flags seg =
	let v = get_field seg fi_flags in
	let i = to_int v in
	let p_x = (i land phf_x_bit) <> 0 in
	let p_w = (i land phf_w_bit) <> 0 in
	let p_r = (i land phf_r_bit) <> 0 in
	{
		phf_x = p_x;
		phf_w = p_w;
		phf_r = p_r;
	}

let get_align seg = get_field seg fi_align

let get_program_header seg =
	let p_type = get_type seg in
	let p_offset = get_offset seg in
	let p_vaddr = get_vaddr seg in
	let p_paddr = get_paddr seg in
	let p_filesz = get_filesz seg in
	let p_memsz = get_memsz seg in
	let p_flags = get_flags seg in
	let p_align = get_align seg in
	{
		p_type = p_type;
		p_offset = p_offset;
		p_vaddr = p_vaddr;
		p_paddr = p_paddr;
		p_filesz = p_filesz;
		p_memsz = p_memsz;
		p_flags = p_flags;
		p_align = p_align;
	}

let get_program_headers header seg =
	let off = to_int header.e_phoff in
	let entsize = to_int header.e_phentsize in
	let num = to_int header.e_phnum in

	get_headers_common seg off entsize num get_program_header

let in_program_header va sz ph =
	let b = va in
	let e = add b (of_int sz) in

	let pb = ph.p_vaddr in
	let pe = add pb ph.p_memsz in

	(pb <= b) && (b <= pe) && (pb < e) && (e <= pe)

let get_segment_seg program_header seg =
        let off = to_int program_header.p_offset in
        let size = to_int program_header.p_memsz in

        get_subseg seg off size

