(*
 * Copyright 2005 Toshiyuki Maeda
 *
 * This file is part of TALK.
 *
 * TALK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * TALK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *)

open Machine

type address_size_attribute =
	| AdSizeAttr16
	| AdSizeAttr32
(*
	| AdSizeAttr64
*)

type operand_size_attribute =
	| OpSizeAttr16
	| OpSizeAttr32
(*
	| OpSizeAttr64
*)

type operand_size =
	| OpSize8
	| OpSize16
	| OpSize32
	| OpSize48
	| OpSize64
	| OpSize128

type prefixes =
{
	g1_opt : prefix_group1 option;
	g2_opt : prefix_group2 option;
	g3_opt : prefix_group3 option;
	g4_opt : prefix_group4 option;
}

and prefix_group1 =
(* lock and repeat prefixes *)
	| LOCK
	| REPNE_REPNZ
	| REP_or_REPE_REPZ

and prefix_group2 =
(* segment override prefixes and branch hints*)
	| WITH_CS_or_BRANCH_NOT_TAKEN
	| WITH_SS
	| WITH_DS_or_BRANCH_TAKEN
	| WITH_ES
	| WITH_FS
	| WITH_GS

and prefix_group3 =
	| OPSIZE_OVERRIDE

and prefix_group4 =
	| ADSIZE_OVERRIDE

and addressing_method =
	| AM_A
	| AM_C
	| AM_D
	| AM_E
	| AM_F
	| AM_G
	| AM_I
	| AM_J
	| AM_M
	| AM_O
	| AM_P
	| AM_Q
	| AM_R
	| AM_S
	| AM_T
	| AM_V
	| AM_W
	| AM_X
	| AM_Y

and operand_type =
	| OT_a
	| OT_b
	| OT_c
	| OT_d
	| OT_dq
	| OT_p
	| OT_pi
	| OT_pd
	| OT_ps
	| OT_q
	| OT_s
	| OT_sd
	| OT_ss
	| OT_si
	| OT_v
	| OT_w

and operand_info = addressing_method * operand_type

and operand =
	| Op_Mem of mem_operand
	| Op_Imd of nativeint * operand_size
	| Op_Reg of register

and mem_scaled_index =
{
	msi_index	: register;
	msi_scale	: int;
}

and mem_operand =
{
	mop_disp_opt		: nativeint option;
	mop_base_opt		: register option;
	mop_scaled_index_opt	: mem_scaled_index option;
	mop_size		: operand_size;
}

and register =
	| Reg8 of register8
	| Reg16 of register16
	| Reg32 of register32
	| Regmm of registermm
	| Regxmm of registerxmm
	| Regseg of segment_register

and register8 =
	| AL
	| CL
	| DL
	| BL
	| AH
	| CH
	| DH
	| BH

and register16 =
	| AX
	| CX
	| DX
	| BX
	| SP
	| BP
	| SI
	| DI

and register32 =
	| EAX
	| ECX
	| EDX
	| EBX
	| ESP
	| EBP
	| ESI
	| EDI

and registermm =
	| MM0
	| MM1
	| MM2
	| MM3
	| MM4
	| MM5
	| MM6
	| MM7

and registerxmm =
	| XMM0
	| XMM1
	| XMM2
	| XMM3
	| XMM4
	| XMM5
	| XMM6
	| XMM7

and segment_register =
	| ES
	| CS
	| SS
	| DS
	| FS
	| GS

and condition_code =
	| CC_O
	| CC_NO
	| CC_ULT
	| CC_UGE
	| CC_Z
	| CC_NZ
	| CC_ULE
	| CC_UGT
	| CC_S
	| CC_NS
	| CC_PE
	| CC_PO
	| CC_SLT
	| CC_SGE
	| CC_SLE
	| CC_SGT

and instruction =
	| Inst_ADD	of operand * operand
	| Inst_OR	of operand * operand
	| Inst_ADC	of operand * operand
	| Inst_SBB	of operand * operand
	| Inst_AND	of operand * operand
	| Inst_SUB	of operand * operand
	| Inst_XOR	of operand * operand
	| Inst_CMP	of operand * operand
	| Inst_BOUND	of operand * operand
	| Inst_ARPL	of operand * operand
	| Inst_IMUL	of operand * operand * operand
	| Inst_TEST	of operand * operand
	| Inst_XCHG	of operand * operand
	| Inst_MOV	of operand * operand
	| Inst_LEA	of operand * operand
	| Inst_LES	of operand * operand
	| Inst_LDS	of operand * operand
	| Inst_POP	of operand

	| Inst_Jcc	of condition_code * operand

	| Inst_INC	of operand
	| Inst_DEC	of operand
	| Inst_PUSH	of operand
	| Inst_PUSHF
	| Inst_POPF

	| Inst_CALLN	of operand
	| Inst_CALLF	of operand
	| Inst_JMPN	of operand
	| Inst_JMPF	of operand
	| Inst_JMPFI	of operand * operand
	| Inst_RET

and instruction_info =
{
	inst_off	: int;
	inst_size	: int;
	inst_opcode	: instruction;
}

let inst_ADD op1 op2	= Inst_ADD (op1, op2)
let inst_OR op1 op2	= Inst_OR (op1, op2)
let inst_ADC op1 op2	= Inst_ADC (op1, op2)
let inst_SBB op1 op2	= Inst_SBB (op1, op2)
let inst_AND op1 op2	= Inst_AND (op1, op2)
let inst_SUB op1 op2	= Inst_SUB (op1, op2)
let inst_XOR op1 op2	= Inst_XOR (op1, op2)
let inst_CMP op1 op2	= Inst_CMP (op1, op2)
let inst_BOUND op1 op2	= Inst_BOUND (op1, op2)
let inst_ARPL op1 op2	= Inst_ARPL (op1, op2)
let inst_IMUL op1 op2 op3	= Inst_IMUL (op1, op2, op3)
let inst_TEST op1 op2	= Inst_TEST (op1, op2)
let inst_XCHG op1 op2	= Inst_XCHG (op1, op2)
let inst_MOV op1 op2	= Inst_MOV (op1, op2)
let inst_LEA op1 op2	= Inst_LEA (op1, op2)
let inst_LES op1 op2	= Inst_LES (op1, op2)
let inst_LDS op1 op2	= Inst_LES (op1, op2)
let inst_POP op		= Inst_POP op

let inst_INC op		= Inst_INC op
let inst_DEC op		= Inst_DEC op
let inst_PUSH op	= Inst_PUSH op

let inst_Jcc cc op	= Inst_Jcc (cc, op)

let inst_CALLN op	= Inst_CALLN op
let inst_CALLF op	= Inst_CALLF op
let inst_JMPN op 	= Inst_JMPN op
let inst_JMPF op 	= Inst_JMPF op
let inst_JMPFI op1 op2 	= Inst_JMPFI (op1, op2)

let oi_Eb = AM_E, OT_b
let oi_Gb = AM_G, OT_b
let oi_Ev = AM_E, OT_v
let oi_Gv = AM_G, OT_v
let oi_Ma = AM_O, OT_a
let oi_Ew = AM_E, OT_w
let oi_Gw = AM_G, OT_w
let oi_Ib = AM_I, OT_b
let oi_Iv = AM_I, OT_v
let oi_Jb = AM_J, OT_b
let oi_Jv = AM_J, OT_v
let oi_Sw = AM_S, OT_w
let oi_M  = AM_M, OT_b	(* oi_M is for 'lea', that is, OT_b is just a place holder. *)
let oi_Mp = AM_M, OT_p
let oi_Ep = AM_E, OT_p
let oi_Ob = AM_O, OT_b
let oi_Ov = AM_O, OT_v

let ot_to_op_size ot op_attr =
	match ot with
	| OT_a  -> if op_attr = OpSizeAttr16 then OpSize32 else OpSize64
	| OT_b  -> OpSize8
	| OT_c  -> if op_attr = OpSizeAttr16 then OpSize8 else OpSize16
	| OT_d  -> OpSize32
	| OT_dq -> OpSize128
	| OT_p  -> if op_attr = OpSizeAttr16 then OpSize32 else OpSize48
	| OT_pi -> OpSize64
	| OT_pd -> OpSize128
	| OT_ps -> OpSize128
	| OT_q  -> OpSize64
	| OT_s  -> OpSize48
	| OT_sd -> OpSize128
	| OT_ss -> OpSize128
	| OT_si -> OpSize32
	| OT_v  -> if op_attr = OpSizeAttr16 then OpSize16 else OpSize32
	| OT_w  -> OpSize16
