(*
 * Copyright 2005 Toshiyuki Maeda
 *
 * This file is part of TALK.
 *
 * TALK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * TALK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *)

open Nativeint
open Bigarray

let bits_per_byte = 8

type segment =
{
	seg_va		: nativeint;

	seg_array	: (int, int8_unsigned_elt, c_layout ) Array1.t;
	seg_offset	: int;
	seg_size	: int;
}

let error_invalid_offset str size off =
	let message = Printf.sprintf "%s: invalid offset %d (size = %d)" str off size in
	raise (Invalid_argument message)

let get_subseg seg off size =
	{ seg with seg_va = add seg.seg_va (of_int off); seg_offset = seg.seg_offset + off; seg_size = size; }

let get_uint_of_size sz seg off =
	if (off + sz > seg.seg_size) then
		error_invalid_offset "Machine.get_int_of_size" seg.seg_size (off + sz - 1);

	let a = seg.seg_array in
	let o = seg.seg_offset in
	let res = ref zero in
	for i = 0 to sz - 1 do
		let b = Array1.get a (i + o + off) in
		let nb = of_int b in
		let shift_bits = i * bits_per_byte in
		res := logor !res (shift_left nb shift_bits)
	done;
	!res

let get_sint_of_size sz seg off =
	let u = get_uint_of_size sz seg off in
	let s = Nativeint.size - sz * bits_per_byte in
	shift_right (shift_left u s) s

let get_byte seg off =
	if (off + 1 > seg.seg_size) then
		error_invalid_offset "Machine.get_byte" seg.seg_size off;

	let a = seg.seg_array in
	let o = seg.seg_offset in
	Array1.get seg.seg_array (o + off)

(* The followings are machine-dependent routines and data *)

let get_uint8 seg off  = get_uint_of_size ( 8 / bits_per_byte) seg off
let get_sint8 seg off  = get_sint_of_size ( 8 / bits_per_byte) seg off
let get_uint16 seg off = get_uint_of_size (16 / bits_per_byte) seg off
let get_sint16 seg off = get_sint_of_size (16 / bits_per_byte) seg off
let get_uint32 seg off = get_uint_of_size (32 / bits_per_byte) seg off
let get_sint32 seg off = get_sint_of_size (32 / bits_per_byte) seg off

let get_word = get_uint32
let size_of_word = 32 / bits_per_byte
