(*
 * Copyright 2005 Toshiyuki Maeda
 *
 * This file is part of TALK.
 *
 * TALK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * TALK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *)

open Nativeint
open Machine
open Read_elf_common
open Util
open Readable_types


let get_char seg off =
	let b = get_byte seg off in
	let c = char_of_int b in
	(off + 1), c

let rec get_uleb128_sub result shift seg off =
	let b = get_byte seg off in
	let off = off + 1 in

	let b6_0 = b land 0x007f in
	let b7   = b land 0x0080 in

	let result = result lor (b6_0 lsr shift) in

	if (b7 = 0) then
		off, result
	else
		get_uleb128_sub result (shift + 7) seg off

let get_uleb128 seg off =
	get_uleb128_sub 0 0 seg off

let rec read_list_sub n read seg off =
	if n = 0 then off, []

	else

	let off, v = read seg off in

	let off, rst = read_list_sub (n - 1) read seg off in

	off, (v :: rst)

let read_list read seg off =
	let off, n = get_uleb128 seg off in

	read_list_sub n read seg off

let read_string seg off =
	let off, len = get_uleb128 seg off in
	let str = String.create len in

	for i = 0 to (len - 1) do
		let _, c = get_char seg (off + i)in
		str.[i] <- c
	done;

	let off = off + len in

	off, str

let read_bool seg off =
	let off, c = get_char seg off in
	match c with
	| 'T' ->
		off, true
	| 'F' ->
		off, false
	| _ ->
		failwithf "Unknown Bool Value '%c'" c

let read_tv seg off =
	let off, s = read_string seg off in
	off, TV(s)

let read_kind seg off =
	let off, c = get_char seg off in
	let k = match c with
	| 'I' ->
		KInt
	| 'K' ->
		KStack
	| 'S' ->
		KStore
	| _ ->
		failwithf "Unknown Kind '%c'" c in
	off, k

let read_tv_and_kind seg off =
	let off, tv = read_tv seg off in
	let off, kind = read_kind seg off in
	off, (tv, kind)

let read_tvs_and_kinds seg off =
	read_list read_tv_and_kind seg off

let read_nativeint seg off =
	let ni = get_word seg off in
	(off + size_of_word), ni

let rec read_integer seg off =
	let off, c = get_char seg off in

	match c with
(*
	| 'C' ->
		let off, ni = read_nativeint seg off in
		let s = Format.sprintf "0x%08nx" ni in
		off, IConst s
*)
	| 'U' ->
		let off, ni = read_nativeint seg off in
		let s = Format.sprintf "0x%08nx" ni in
		off, IConst s
	| 'S' ->
		let off, ni = read_nativeint seg off in
		let s = Format.sprintf "%nd" ni in
		off, IConst s
	| 'V' ->
		let off, tv = read_tv seg off in
		off, IVar tv
	| _ ->
		failwithf "Unknown Integer type '%c'." c

let read_arith_op seg off =
	let off, c = get_char seg off in

	match c with
	| '+' ->
		off, A_ADD
	| '*' ->
		off, A_MUL
	| _ ->
		failwithf "Unknown Arithmetic Operator '%c'." c


let read_cmp_op_raw seg off =
	let off, c = get_char seg off in

	match c with
	| '=' ->
		off, C_EQ
	| 'N' ->
		off, C_NEQ
	| '<' ->
		off, C_LT
	| '>' ->
		off, C_GT
	| 'L' ->
		off, C_LE
	| 'G' ->
		off, C_GE
	| _ ->
		failwithf "Unknown Comparison Operator '%c'." c

let read_cmp_op seg off =
	let off, cop_r = read_cmp_op_raw seg off in
	let off, signed = read_bool seg off in
	off, (cop_r, signed)

let rec read_integer_expr seg off =
	let off, c = get_char seg off in

	match c with
	| 'I' ->
		let off, inte = read_integer seg off in
		off, IEInte inte
	| 'A' ->
		let off, aop = read_arith_op seg off in
		let off, i_e1 = read_integer_expr seg off in
		let off, i_e2 = read_integer_expr seg off in
		off, IEArith (aop, i_e1, i_e2)
	| _ ->
		failwithf "Unknown Integer Expression type '%c'." c

let read_integer_cstrt seg off =
	let off, c = get_char seg off in

	match c with
	| 'C' ->
		let off, cop = read_cmp_op seg off in
		let off, i_e1 = read_integer_expr seg off in
		let off, i_e2 = read_integer_expr seg off in
		off, ICCmp (cop, i_e1, i_e2)
	| _ ->
		failwithf "Unknown Integer Constraint type '%c'." c

let read_integer_cstrts seg off =
	read_list read_integer_cstrt seg off

let read_register_kind seg off =
	let off, c = get_char seg off in
	let reg_kind = match c with
		| 'L' -> RK_L
		| 'H' -> RK_H
		| 'X' -> RK_X
		| 'E' -> RK_E
		| _ ->
			failwithf "Unknown register kind '%c'." c
	in
	off, reg_kind

let read_register seg off =
	let off, reg_kind = read_register_kind seg off in
	let off, inte = read_integer seg off in
	off, (reg_kind, inte)

let read_operand_size seg off =
	let off, c = get_char seg off in
	let op_size = match c with
		| '1' -> Instructions.OpSize8
		| '2' -> Instructions.OpSize16
		| '4' -> Instructions.OpSize32
		| '6' -> Instructions.OpSize48
		| '8' -> Instructions.OpSize64
		| '0' -> Instructions.OpSize128
		| _ ->
			failwithf "Unknown operand size '%c'." c
	in
	off, op_size

let read_flags seg off =
	let off, f_rst = read_integer seg off in
	let off, f_op1 = read_integer seg off in
	let off, f_op2 = read_integer seg off in
	let off, f_size = read_operand_size seg off in

	let flags = { f_rst = f_rst; f_op1 = f_op1; f_op2 = f_op2; f_size = f_size; } in

	off, flags

let read_flags_opt seg off =
	let off, c = get_char seg off in
	match c with
	| 'N' ->
		off, None
	| 'S' ->
		let off, flags = read_flags seg off in
		off, Some flags
	| _ ->
		failwithf "Unknown flags option '%c'." c

let read_registers seg off =
	let off, eax = read_register seg off in
	let off, ecx = read_register seg off in
	let off, edx = read_register seg off in
	let off, ebx = read_register seg off in
	let off, esp = read_register seg off in
	let off, ebp = read_register seg off in
	let off, esi = read_register seg off in
	let off, edi = read_register seg off in

	let off, flags_opt = read_flags_opt seg off in

	let regs =
	{
		cs = CodeSeg (SegSize32);
		eax = eax;
		ecx = ecx;
		edx = edx;
		ebx = ebx;
		esp = esp;
		ebp = ebp;
		esi = esi;
		edi = edi;
		flags_opt = flags_opt;
	} in

	off, regs

let rec read_stack_type seg off =
	let off, c = get_char seg off in
	match c with
	| 'N' ->
		off, StkNull
	| 'V' ->
		let off, tv = read_tv seg off in
		off, StkVar tv
	| 'C' ->
		let off, inte = read_integer seg off in
		let off, st = read_stack_type seg off in
		off, StkCons (inte, st)
	| _ ->
		failwithf "Unknown Stack type '%c'." c

let read_rich_stack_type seg off =
	let off, freelen = read_integer seg off in
	let off, stk = read_stack_type seg off in

	off, { rstk_stk = stk; rstk_freelen = freelen; }

let read_tuple_elem_type seg off =
	let off, op_size = read_operand_size seg off in
	let off, inte = read_integer seg off in
	off, (op_size, inte)

let read_tuple_type seg off =
	read_list read_tuple_elem_type seg off

let rec read_atom seg off =
	let off, inte = read_integer seg off in
	let off, ht = read_heap_type seg off in
	off, (inte, ht)

and read_code_atom seg off =
	let off, inte = read_integer seg off in
	let off, ct = read_code_type seg off in
	off, (inte, ct)

and read_store_elem seg off =
	let off, c = get_char seg off in
	match c with
	| 'A' ->
		let off, a = read_atom seg off in
		off, SEAtom a
	| 'V' ->
		let off, tv = read_tv seg off in
		off, SEVar tv
	| _ ->
		failwithf "Unknown Store '%c'." c

and read_store seg off =
	read_list read_store_elem seg off

and read_code_store seg off =
	read_list read_code_atom seg off

and read_cstrt seg off =
	let off, ics = read_integer_cstrts seg off in
	let off, store = read_store seg off in
	let off, c_str = read_code_store seg off in
	off, { c_ics = ics; c_store = store; c_code_store = c_str; }

and read_prefix seg off =
	let off, tvs_and_kinds = read_tvs_and_kinds seg off in
	let off, cstrts = read_list read_cstrt seg off in
	off, { p_tvs = tvs_and_kinds; p_cstrts = cstrts; }

and read_state_type seg off =
	let off, prefix = read_prefix seg off in
	let off, regs = read_registers seg off in

	off, { st_prefix = prefix; st_registers = regs; }

and read_pack_type seg off =
	let off, prefix = read_prefix seg off in
	let off, tt = read_tuple_type seg off in

	off, { pt_prefix = prefix; pt_tuple = tt; }

and read_concrete_type seg off =
	let off, c = get_char seg off in
	match c with
	| 'P' ->
		let off, pt = read_pack_type seg off in
		off, Pack pt
	| 'T' ->
		let off, tt = read_tuple_type seg off in
		off, Tuple tt
	| _ ->
		failwithf "Unknown Concrete type '%c'." c

and read_type seg off =
	let off, c = get_char seg off in
	match c with
	| 'C' ->
		let off, ct = read_concrete_type seg off in
		off, Concrete ct
	| 'N' ->
		let off, nt = read_named_type seg off in
		off, Name nt
	| _ ->
		failwithf "Unknown Type '%c'." c

and read_array_type seg off =
	let off, typ = read_type seg off in
	let off, len = read_integer seg off in

	off, { at_type = typ; at_len = len; }

and read_heap_type seg off =
	let off, c = get_char seg off in
	match c with
	| 'A' ->
		let off, at = read_array_type seg off in
		off, Array at
	| 'S' ->
		let off, rstk = read_rich_stack_type seg off in
		off, Stack rstk
	| _ ->
		failwithf "Unkonwo Heap Type '%c'." c


and read_code_type seg off =
	let off, c = get_char seg off in
	match c with
	| 'S' ->
		let off, st = read_state_type seg off in
		off, State st
	| 'N' ->
		let off, nt = read_named_type seg off in
		off, NameCode nt
	| _ ->
		failwithf "Unknown Code Type '%c'." c

and read_named_type seg off =
	let off, name = read_string seg off in
	let off, cs = read_constructors seg off in
	off, (name, cs)

and read_constructor seg off =
	let off, c = get_char seg off in
	match c with
	| 'I' ->
		let off, inte = read_integer seg off in
		off, CInt inte
	| 'K' ->
		let off, st = read_stack_type seg off in
		off, CStack st
	| 'S' ->
		let off, st = read_store seg off in
		off, CStore st
	| _ ->
		failwithf "Unknown Constructor '%c'." c

and read_constructors seg off =
	read_list read_constructor seg off

let read_named_type_body seg off =
	let off, c = get_char seg off in
	match c with
	| 'C' ->
		let off, ct = read_concrete_type seg off in
		off, NtdConcrete ct
	| 'S' ->
		let off, st = read_state_type seg off in
		off, NtdState st
	| _ ->
		failwithf "Unknown Name Type Body '%c'." c

let read_named_type_definition seg off =
	let off, tvs_and_kinds = read_tvs_and_kinds seg off in
	let off, bdy = read_named_type_body seg off in

	off, { ntd_tvs = tvs_and_kinds; ntd_body = bdy; }
