(*
 * Copyright 2005 Toshiyuki Maeda
 *
 * This file is part of TALK.
 *
 * TALK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * TALK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *)

(* De Bruijn's index! *)

(********************************
	Type definition
	      for
	module definitions
 ********************************)

type type_variable = TV of int

module TTV =
struct
	type t = type_variable
	let compare = Pervasives.compare
end

module TVSet = Set.Make (TTV)

(********************************
	Type definition
	      for
	integer constraints
 ********************************)

type integer =
	| IConst of Big_int.big_int
	| IVar of type_variable

type arith_op =
	| A_ADD
	| A_MUL

type cmp_op_raw =
	| C_EQ
	| C_NEQ
	| C_LT
	| C_GT
	| C_LE
	| C_GE

type cmp_op = cmp_op_raw * bool

type integer_expr =
	| IEInte of integer
	| IEArith of arith_op * integer_expr * integer_expr

type integer_cstrt =
	| ICCmp of cmp_op * integer_expr * integer_expr

type integer_cstrts = integer_cstrt list 
(*
 * NOTE: integer constraints variables are unneccessary,
 * because the unifier and constraint solver do the job.
 *)

module TInt =
struct
	type t = integer
	let compare i1 i2 =
		match i1, i2 with
		| IConst bi1, IConst bi2 ->
			Big_int.compare_big_int bi1 bi2
		| IVar tv1, IVar tv2 ->
			Pervasives.compare tv1 tv2
		| IConst _, IVar _ ->
			-1
		| IVar _, IConst _ ->
			1
end

module IntMap = Map.Make (TInt)



(********************************
	Type definitions
 ********************************)

type kind =
	| KStack
	| KInt
	| KStore

type kinds = (kind * string) list

type typ =
	| Name of named_type
	| Concrete of concrete_type

and array_type =
{
	at_type	: typ;
	at_len	: integer;
}

and stack_type =
	| StkNull
	(* [] *)

	| StkVar of type_variable

	| StkCons of integer * stack_type
	(* wt :: stk *)

and rich_stack_type =
{
	rstk_stk	: stack_type;
(* is this required?
	rstk_size	: integer;
*)
	rstk_freelen	: integer;
}

and code_type =
	| NameCode of named_type
	| State of state_type

and named_type = string * (constructor list)

and state_type =
{
	st_prefix	: prefix;
	st_registers	: registers_type;
}

and concrete_type =
	| Pack of pack_type
	(* Exists[a1:k1, ..., an:kn | constraints].rt *)

	| Tuple of tuple_type

and prefix =
{
	(* De Bruijn's notation! *)
	(*
	 * L:KWord . L:KStack . L:KInt . L:KStore. ===> [KStore; KInt; KStack; KWord; ]
	 *)
	p_kinds		: kinds;
	(* string == original name of the type variable *)

	p_cstrts	: cstrt list;
}

and cstrt =
{
	c_ics		: integer_cstrts;
	c_store		: store;
	c_code_store	: code_store;
}

and pack_type =
{
	pt_prefix	: prefix;
	pt_tuple	: tuple_type;
}

(*
and tuple_type = integer list
*)

and tuple_type = tuple_elem_type list

and tuple_elem_type = Instructions.operand_size * integer

and registers_type =
{
	cs	: segment_type;

(*
	eax	: integer;
	ecx	: integer;
	edx	: integer;
	ebx	: integer;
	esi	: integer;
	edi	: integer;
	ebp	: integer;
	esp	: integer;

	frt	: integer;
	fo1	: integer;
	fo2	: integer;
*)
	eax	: register_type;
	ecx	: register_type;
	edx	: register_type;
	ebx	: register_type;
	esi	: register_type;
	edi	: register_type;
	ebp	: register_type;
	esp	: register_type;

	flags_opt	: flags_type option;
}

and flags_type =
{
	f_rst	: integer;
	f_op1	: integer;
	f_op2	: integer;
	f_size	: Instructions.operand_size;
}

and segment_size_type =
	| SegSize16
	| SegSize32

and segment_type =
	| CodeSeg of segment_size_type
	| DataSeg of segment_size_type

and register_kind =
	| RK_L
	| RK_H
	| RK_X
	| RK_E

and register_type = register_kind * integer

and heap_type =
	| Array of array_type
	| Stack of rich_stack_type

and store =
{
	str_locs	:	heap_type IntMap.t;
	str_vs		:	TVSet.t;
}

and code_store = code_type IntMap.t

and constructor =
	| CStack of stack_type
	| CInt of integer
	| CStore of store

type named_type_body =
	| NtdConcrete of concrete_type
	| NtdState of state_type

type named_type_definition =
{
	ntd_kinds	: kinds;
	ntd_body	: named_type_body;
}

(*
 * Type checker context
 *)
type context =
{
	ctx_kinds : kinds;
	ctx_cstrt : cstrt;
	ctx_registers : registers_type;
}

