(*
 * Copyright 2005 Toshiyuki Maeda
 *
 * This file is part of TALK.
 *
 * TALK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * TALK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *)

open Types

open Big_int_util

(************************************
	Store manipulation (part1)
 ************************************)

let forall f m =
try
	IntMap.iter (fun k v ->
		if not(f k v) then
			raise Not_found
	) m;
	true
with _ ->
	false

exception Invalid_store_manipulation

let empty_integer_cstrts = []

let empty_store =
{
	str_locs = IntMap.empty;
	str_vs = TVSet.empty;
}

let empty_code_store = IntMap.empty

let is_empty_store str =
	IntMap.is_empty str.str_locs && TVSet.is_empty str.str_vs

let is_empty_code_store c_str =
	IntMap.is_empty c_str

let replace_location inte at locs =
	if IntMap.mem inte locs then
		IntMap.add inte at locs
	else
		raise Invalid_store_manipulation

let replace_store_location inte at str =
	let locs = str.str_locs in
	let locs = replace_location inte at locs in
	{ str with str_locs = locs; }

let add_location inte at locs =
	if IntMap.mem inte locs then
		raise Invalid_store_manipulation
	else
		IntMap.add inte at locs

let add_code_location inte st c_str =
	if IntMap.mem inte c_str then
		raise Invalid_store_manipulation
	else
		IntMap.add inte st c_str

let remove_location inte locs =
	if IntMap.mem inte locs then
		IntMap.remove inte locs
	else
		raise Invalid_store_manipulation

let add_store_location inte at str =
	let locs = str.str_locs in
	let locs = add_location inte at locs in
	{ str with str_locs = locs; }

let remove_store_location inte str =
	let locs = str.str_locs in
	let locs = remove_location inte locs in
	{ str with str_locs = locs; }

let add_variable tv vs =
	if TVSet.mem tv vs then
		raise Invalid_store_manipulation
	else
		TVSet.add tv vs

let add_store_variable tv str =
	let vs = str.str_vs in
	let vs = add_variable tv vs in
	{ str with str_vs = vs; }

let add_locations locs1 locs2 =
	IntMap.fold (fun inte at new_locs ->
		add_location inte at new_locs
	) locs1 locs2

let add_variables vs1 vs2 =
	TVSet.fold (fun tv new_vs ->
		add_variable tv new_vs
	) vs1 vs2

let add_store str1 str2 =
	let locs1 = str1.str_locs in
	let vs1 = str1.str_vs in

	let locs2 = str2.str_locs in
	let vs2 = str2.str_vs in

	let locs = add_locations locs1 locs2 in
	let vs = add_variables vs1 vs2 in

	{ str_locs = locs; str_vs = vs; }

let add_code_store c_str1 c_str2 =
	IntMap.fold (fun inte st new_c_str ->
		add_code_location inte st new_c_str
	) c_str1 c_str2

(************************************
	Integer Constraints
 ************************************)

let add_integer_cstrt ic ics =
	ic :: ics

let add_integer_cstrts ics1 ics2 =
	ics1 @ ics2


(************************************
	Type Equality Check
 ************************************)

let is_equal_type_variable tv1 tv2 =
	let TV(i1) = tv1 in
	let TV(i2) = tv2 in
	i1 = i2

open Integer_op

let is_equal_integer ics i1 i2 =
	match (i1, i2) with
	| (IConst n1, IConst n2) ->
		n1 =^ n2
	| (IVar tv1, IVar tv2) when is_equal_type_variable tv1 tv2 ->
		true	
	| _ ->
		let ie1 = ie_inte i1 in
		let ie2 = ie_inte i2 in
		let ic = (ie1 =.. ie2) in
		Constraint_solver.implies_one ics ic

(*
let is_equal_integer_cstrts ics ics1 ics2 =
	let b2 = Constraint_solver.implies (add_integer_cstrts ics ics1) ics2 in
	let b1 = Constraint_solver.implies (add_integer_cstrts ics ics2) ics1 in
	b1 && b2
*)
let is_equal_integer_cstrts ics ics1 ics2 =
	let b2 = Constraint_solver.implies ics1 ics2 in
	let b1 = Constraint_solver.implies ics2 ics1 in
	b1 && b2

let is_equal_tuple_elem_type ics (sz1, inte1) (sz2, inte2) =
	sz1 = sz2 &&
	is_equal_integer ics inte1 inte2

let is_equal_tuple_type ics tt1 tt2 =
	if List.length tt1 <> List.length tt2 then
		false
	else
		List.for_all2 (is_equal_tuple_elem_type ics) tt1 tt2

let rec is_equal_stack_type ics st1 st2 =
	match (st1, st2) with
	| StkNull, StkNull ->
		true
	| StkVar tv1, StkVar tv2 ->
		is_equal_type_variable tv1 tv2
	| StkCons (inte1, st1), StkCons(inte2, st2) ->
		is_equal_integer ics inte1 inte2 &&
		is_equal_stack_type ics st1 st2
	| _ ->
		false

let is_equal_rich_stack_type ics rstk1 rstk2 =
	let stk1 = rstk1.rstk_stk in
	let stk2 = rstk2.rstk_stk in

	let frlen1 = rstk1.rstk_freelen in
	let frlen2 = rstk2.rstk_freelen in

	is_equal_stack_type ics stk1 stk2 &&
	is_equal_integer ics frlen1 frlen2

let rec is_equal_cstrt ics cstrt1 cstrt2 =
	let ics1 = cstrt1.c_ics in
	let ics2 = cstrt2.c_ics in

	let store1 = cstrt1.c_store in
	let store2 = cstrt2.c_store in

	let c_str1 = cstrt1.c_code_store in
	let c_str2 = cstrt2.c_code_store in

	is_equal_store ics store1 store2
	&& is_equal_code_store ics c_str1 c_str2
	&& is_equal_integer_cstrts ics ics1 ics2

and is_equal_cstrts ics cstrts1 cstrts2 =
	Util.is_equal_unordered (is_equal_cstrt ics) cstrts1 cstrts2

and is_equal_prefix ics prefix1 prefix2 =
	let kinds1, _ = List.split prefix1.p_kinds in
	let kinds2, _ = List.split prefix2.p_kinds in

	let cstrts1 = prefix1.p_cstrts in
	let cstrts2 = prefix2.p_cstrts in

	kinds1 = kinds2 && is_equal_cstrts ics cstrts1 cstrts2
	
and is_equal_concrete_type ics ct1 ct2 =
	match (ct1, ct2) with
	| (Pack pt1, Pack pt2) ->
		is_equal_pack_type ics pt1 pt2
	| (Tuple tt1, Tuple tt2) ->
		is_equal_tuple_type ics tt1 tt2
	| _ ->
		false

and is_equal_state_type ics st1 st2 =
	let { st_prefix = prefix1; st_registers = regs1; } = st1 in
	let { st_prefix = prefix2; st_registers = regs2; } = st2 in
	is_equal_prefix ics prefix1 prefix2 && is_equal_registers_type ics regs1 regs2

and is_equal_pack_type ics pt1 pt2 =
	let { pt_prefix = prefix1; pt_tuple = tt1; } = pt1 in
	let { pt_prefix = prefix2; pt_tuple = tt2; } = pt2 in
	is_equal_prefix ics prefix1 prefix2 && is_equal_tuple_type ics tt1 tt2

and is_equal_registers_type ics regs1 regs2 =
	is_equal_register_type ics regs1.eax regs2.eax &&
	is_equal_register_type ics regs1.ecx regs2.ecx &&
	is_equal_register_type ics regs1.edx regs2.edx &&
	is_equal_register_type ics regs1.ebx regs2.ebx &&
	is_equal_register_type ics regs1.esi regs2.esi &&
	is_equal_register_type ics regs1.edi regs2.edi &&
	is_equal_register_type ics regs1.ebp regs2.ebp &&
	is_equal_register_type ics regs1.esp regs2.esp &&
	is_equal_flags_type_opt ics regs1.flags_opt regs2.flags_opt

and is_equal_flags_type_opt ics flags_opt1 flags_opt2 =
	match flags_opt1, flags_opt2 with
	| Some flags1, Some flags2 ->
		is_equal_flags_type ics flags1 flags2
	| None, None ->
		true
	| _ ->
		false

and is_equal_flags_type ics flags1 flags2 =
	flags1.f_size = flags2.f_size &&
	is_equal_integer ics flags1.f_rst flags2.f_rst &&
	is_equal_integer ics flags1.f_op1 flags2.f_op1 &&
	is_equal_integer ics flags1.f_op2 flags2.f_op2

and is_equal_register_kind reg_kind1 reg_kind2 =
	match reg_kind1, reg_kind2 with
	| RK_L, RK_L | RK_H, RK_H | RK_X, RK_X | RK_E, RK_E ->
		true
	| _ ->
		false

and is_equal_register_type ics (reg_kind1, inte1) (reg_kind2, inte2) =
	is_equal_register_kind reg_kind1 reg_kind2 &&
	is_equal_integer ics inte1 inte2
	
and is_equal_store ics str1 str2 =
	let locs1 = str1.str_locs in
	let locs2 = str2.str_locs in
	let vs1 = str1.str_vs in
	let vs2 = str2.str_vs in

	IntMap.equal (is_equal_heap_type ics) locs1 locs2
	&& TVSet.equal vs1 vs2

and is_equal_code_store ics c_str1 c_str2 =
	IntMap.equal (is_equal_code_type ics) c_str1 c_str2

and is_equal_constructor ics c1 c2 =
	match (c1, c2) with
	| (CStack st1, CStack st2) ->
		is_equal_stack_type ics st1 st2
	| (CInt inte1, CInt inte2) ->
		is_equal_integer ics inte1 inte2
	| (CStore str1, CStore str2) ->
		is_equal_store ics str1 str2
	| _ ->
		false

and is_equal_type ics t1 t2 =
	match (t1, t2) with
	| Name nt1, Name nt2 ->
		is_equal_named_type ics nt1 nt2
	| Concrete ct1, Concrete ct2 ->
		is_equal_concrete_type ics ct1 ct2
	| _ ->
		false

and is_equal_heap_type ics ht1 ht2 =
	match ht1, ht2 with
	| Array at1, Array at2 ->
		is_equal_array_type ics at1 at2
	| Stack st1, Stack st2 ->
		is_equal_rich_stack_type ics st1 st2
	| _ ->
		false

and is_equal_array_type ics at1 at2 =
	let { at_type = t1; at_len = len1; } = at1 in
	let { at_type = t2; at_len = len2; } = at2 in

	is_equal_type ics t1 t2 && is_equal_integer ics len1 len2

and is_equal_code_type ics ct1 ct2 =
	match (ct1, ct2) with
	| NameCode nt1, NameCode nt2 ->
		is_equal_named_type ics nt1 nt2
	| State st1, State st2 ->
		is_equal_state_type ics st1 st2
	| _ ->
		false

and is_equal_named_type ics nt1 nt2 =
	let n1, cs1 = nt1 in
	let n2, cs2 = nt2 in
	try
		(n1 = n2) &&
		(List.for_all2 (is_equal_constructor ics) cs1 cs2)
	with Invalid_argument _ ->
		false


(************************************
	De Bruijn's shifting
 ************************************)

exception Invalid_shift

let shift_type_variable c d tv =
	let TV(i) = tv in
	let i =
		if i < c then i
		else i + d in
	if i < 0 then
		raise Invalid_shift
	else
		TV(i)

let shift_variables c d vs =
	TVSet.fold (fun tv new_vs ->
		let tv = shift_type_variable c d tv in
		add_variable tv new_vs
	) vs TVSet.empty

let shift_integer c d inte =
	match inte with
	| IVar tv ->
		let tv = shift_type_variable c d tv in
		IVar(tv)
	| _ -> inte

let rec shift_integer_expr c d ie =
	match ie with
	| IEInte inte ->
		let inte = shift_integer c d inte in
		IEInte inte
	| IEArith (aop, ie1, ie2) ->
		let ie1 = shift_integer_expr c d ie1 in
		let ie2 = shift_integer_expr c d ie2 in
		IEArith (aop, ie1, ie2)

let rec shift_integer_cstrt c d ic =
	match ic with
	| ICCmp (cop, ie1, ie2) ->
		let ie1 = shift_integer_expr c d ie1 in
		let ie2 = shift_integer_expr c d ie2 in
		ICCmp (cop, ie1, ie2)

let rec shift_integer_cstrts c d ics =
	List.map (shift_integer_cstrt c d) ics

let shift_tuple_elem_type c d (sz, inte) =
	let inte = shift_integer c d inte in
	(sz, inte)

let shift_tuple_type c d tt =
	List.map (shift_tuple_elem_type c d) tt

let shift_register_type c d reg =
	let (reg_kind, inte) = reg in
	let inte = shift_integer c d inte in
	(reg_kind, inte)

let shift_flags_type c d flags =
{
	f_rst = shift_integer c d flags.f_rst;
	f_op1 = shift_integer c d flags.f_op1;
	f_op2 = shift_integer c d flags.f_op2;
	f_size = flags.f_size;
}

let shift_flags_type_opt c d = function
	| None ->
		None
	| Some flags ->
		let flags = shift_flags_type c d flags in
		Some flags

let shift_registers_type c d regs =
{
(*
	eax = shift_integer c d regs.eax;
	ecx = shift_integer c d regs.ecx;
	edx = shift_integer c d regs.edx;
	ebx = shift_integer c d regs.ebx;
	esi = shift_integer c d regs.esi;
	edi = shift_integer c d regs.edi;
	ebp = shift_integer c d regs.ebp;
	esp = shift_integer c d regs.esp;
	frt = shift_integer c d regs.frt;
	fo1 = shift_integer c d regs.fo1;
	fo2 = shift_integer c d regs.fo2;
*)
	cs = regs.cs;
	eax = shift_register_type c d regs.eax;
	ecx = shift_register_type c d regs.ecx;
	edx = shift_register_type c d regs.edx;
	ebx = shift_register_type c d regs.ebx;
	esi = shift_register_type c d regs.esi;
	edi = shift_register_type c d regs.edi;
	ebp = shift_register_type c d regs.ebp;
	esp = shift_register_type c d regs.esp;

	flags_opt = shift_flags_type_opt c d regs.flags_opt;
}

let rec shift_stack_type c d st =
	match st with
	| StkNull ->
		StkNull
	| StkVar tv ->
		let tv = shift_type_variable c d tv in
		StkVar tv
	| StkCons (inte, st) ->
		let inte = shift_integer c d inte in
		let st = shift_stack_type c d st in
		StkCons (inte, st)

let shift_rich_stack_type c d rstk =
	let stk = shift_stack_type c d rstk.rstk_stk in
	let freelen = shift_integer c d rstk.rstk_freelen in

	{ rstk_stk = stk; rstk_freelen = freelen; }

let rec shift_state_type c d st =
	let (c, prefix) = shift_prefix c d st.st_prefix in
	let regs = shift_registers_type c d st.st_registers in

	{ st_prefix = prefix; st_registers = regs; }

and shift_concrete_type c d ct =
	match ct with
	| Pack pt ->
		let pt = shift_pack_type c d pt in
		Pack pt
	| Tuple tt ->
		let tt = shift_tuple_type c d tt in
		Tuple tt

and shift_pack_type c d pt =
	let (c, prefix) = shift_prefix c d pt.pt_prefix in
	let tt = shift_tuple_type c d pt.pt_tuple in
	{ pt_prefix = prefix; pt_tuple = tt; }

and shift_store c d str =
	let locs = str.str_locs in
	let vs = str.str_vs in

	let locs = shift_locations c d locs in
	let vs = shift_variables c d vs in

	{ str_locs = locs; str_vs = vs; }

and shift_locations c d locs =
	IntMap.fold (fun inte ht new_locs ->
		let inte = shift_integer c d inte in
		let ht = shift_heap_type c d ht in
		add_location inte ht new_locs
	) locs IntMap.empty

and shift_code_store c d c_str =
	IntMap.fold (fun inte st new_c_str ->
		let inte = shift_integer c d inte in
		let st = shift_code_type c d st in
		add_code_location inte st new_c_str
	) c_str IntMap.empty

and shift_prefix c d prefix =
	let len = List.length prefix.p_kinds in
	let c = c + len in

	let cstrts = shift_cstrts c d prefix.p_cstrts in

	let prefix = { prefix with p_cstrts = cstrts; } in
	(c, prefix)

and shift_cstrts c d cstrts =
	List.map (shift_cstrt c d) cstrts

and shift_cstrt c d cstrt =
	let ics = shift_integer_cstrts c d cstrt.c_ics in
	let store = shift_store c d cstrt.c_store in
	let code_store = shift_code_store c d cstrt.c_code_store in

	let cstrt = { c_ics = ics; c_store = store; c_code_store = code_store; } in
	cstrt

and shift_type c d t =
	match t with
	| Name nt ->
		let nt = shift_named_type c d nt in
		Name nt
	| Concrete ct ->
		let ct = shift_concrete_type c d ct in
		Concrete ct

and shift_heap_type c d = function
	| Array at ->
		let at = shift_array_type c d at in
		Array at
	| Stack st ->
		let st = shift_rich_stack_type c d st in
		Stack st

and shift_array_type c d at =
	let t = at.at_type in
	let len = at.at_len in

	let t = shift_type c d t in
	let len = shift_integer c d len in

	{ at_type = t; at_len = len; }

and shift_code_type c d ct =
	match ct with
	| NameCode nt ->
		let nt = shift_named_type c d nt in
		NameCode nt
	| State st ->
		let st = shift_state_type c d st in
		State st

and shift_named_type c d nt =
	let n, cs = nt in
	let cs = List.map (shift_constructor c d) cs in
	(n, cs)

and shift_constructor c d = function
	| CStack st ->
		let st = shift_stack_type c d st in
		CStack st
	| CInt inte ->
		let inte = shift_integer c d inte in
		CInt inte
	| CStore str ->
		let str = shift_store c d str in
		CStore str

let shift_named_type_body c d bdy =
	match bdy with
	| NtdConcrete ct ->
		let ct = shift_concrete_type c d ct in
		NtdConcrete ct
	| NtdState st ->
		let st = shift_state_type c d st in
		NtdState st

let shift_named_type_definition c d ntd =
	let len = List.length ntd.ntd_kinds in
	let c = c + len in

	let bdy = shift_named_type_body c d ntd.ntd_body in

	{ ntd with ntd_body = bdy; }


(************************************
	Anytype operations
 ************************************)
exception Kind_mismatch_error of kind * constructor

let cast_to_stack_type c =
	match c with
	| CStack st -> st
	| _ -> raise (Kind_mismatch_error (KStack, c))

let cast_to_integer c =
	match c with
	| CInt inte -> inte
	| _ -> raise (Kind_mismatch_error (KInt, c))

let cast_to_store c =
	match c with
	| CStore str -> str
	| _ -> raise (Kind_mismatch_error (KStore, c))

(************************************
	De Bruijn's substitution
 ************************************)

let subst_integer tv1 c inte =
	match inte with
	| IVar tv2 ->
		if is_equal_type_variable tv1 tv2 then
			cast_to_integer c 
		else
			inte
	| _ ->
		inte

let rec subst_integer_expr tv c i_e =
	match i_e with
	| IEInte inte ->
		let inte = subst_integer tv c inte in
		IEInte inte
	| IEArith (aop, i_e1, i_e2) ->
		let i_e1 = subst_integer_expr tv c i_e1 in
		let i_e2 = subst_integer_expr tv c i_e2 in
		IEArith (aop, i_e1, i_e2)

let subst_integer_cstrt tv c i_c =
	match i_c with
	| ICCmp (cop, i_e1, i_e2) ->
		let i_e1 = subst_integer_expr tv c i_e1 in
		let i_e2 = subst_integer_expr tv c i_e2 in
		ICCmp (cop, i_e1, i_e2)

let subst_integer_cstrts tv c ics =
	List.map (subst_integer_cstrt tv c) ics

let subst_tuple_elem_type tv c (sz, inte) =
	let inte = subst_integer tv c inte in
	(sz, inte)

let subst_tuple_type tv c tt =
	List.map (subst_tuple_elem_type tv c) tt

let subst_register tv c reg =
	let (reg_kind, inte) = reg in
	let inte = subst_integer tv c inte in
	(reg_kind, inte)

let subst_flags_type tv c flags =
{
	f_rst = subst_integer tv c flags.f_rst;
	f_op1 = subst_integer tv c flags.f_op1;
	f_op2 = subst_integer tv c flags.f_op2;
	f_size = flags.f_size;
}

let subst_flags_type_opt tv c = function
	| None ->
		None
	| Some flags ->
		let flags = subst_flags_type tv c flags in
		Some flags

let subst_registers tv c regs =
{
(*
	eax = subst_integer tv c regs.eax;
	ecx = subst_integer tv c regs.ecx;
	edx = subst_integer tv c regs.edx;
	ebx = subst_integer tv c regs.ebx;
	esi = subst_integer tv c regs.esi;
	edi = subst_integer tv c regs.edi;
	ebp = subst_integer tv c regs.ebp;
	esp = subst_integer tv c regs.esp;
	frt = subst_integer tv c regs.frt;
	fo1 = subst_integer tv c regs.fo1;
	fo2 = subst_integer tv c regs.fo2;
*)
	cs = regs.cs;
	eax = subst_register tv c regs.eax;
	ecx = subst_register tv c regs.ecx;
	edx = subst_register tv c regs.edx;
	ebx = subst_register tv c regs.ebx;
	esi = subst_register tv c regs.esi;
	edi = subst_register tv c regs.edi;
	ebp = subst_register tv c regs.ebp;
	esp = subst_register tv c regs.esp;

	flags_opt = subst_flags_type_opt tv c regs.flags_opt;
}

let rec subst_stack_type tv1 c st =
	match st with
	| StkNull ->
		StkNull
	| StkVar tv2 ->
		if is_equal_type_variable tv1 tv2 then
			cast_to_stack_type c
		else
			st
	| StkCons (inte, st) ->
		let inte = subst_integer tv1 c inte in
		let st = subst_stack_type tv1 c st in
		StkCons (inte, st)

let subst_rich_stack_type tv c rstk =
	let stk = subst_stack_type tv c rstk.rstk_stk in
	let freelen = subst_integer tv c rstk.rstk_freelen in

	{ rstk_stk = stk; rstk_freelen = freelen; }

let rec subst_state_type tv c st =
	let (tv, c, prefix) = subst_prefix tv c st.st_prefix in
	let regs = subst_registers tv c st.st_registers in

	{ st_prefix = prefix; st_registers = regs; }

and subst_concrete_type tv c ct =
	match ct with
	| Pack pt ->
		let pt = subst_pack_type tv c pt in
		Pack pt
	| Tuple tt ->
		let tt = subst_tuple_type tv c tt in
		Tuple tt

and subst_pack_type tv c pt =
	let (tv, c, prefix) = subst_prefix tv c pt.pt_prefix in
	let tt = subst_tuple_type tv c pt.pt_tuple in
	{ pt_prefix = prefix; pt_tuple = tt; }

and subst_store tv c str =

	let locs = str.str_locs in
	let locs = subst_locations tv c locs in

	let vs = str.str_vs in
	let str = subst_variables tv c vs in
	let locs' = str.str_locs in
	let vs = str.str_vs in

	let locs = add_locations locs locs' in

	{ str_locs = locs; str_vs = vs; }

and subst_locations tv c locs =
	IntMap.fold (fun inte ht new_locs ->
		let inte = subst_integer tv c inte in
		let ht = subst_heap_type tv c ht in
		add_location inte ht new_locs
	) locs IntMap.empty

and subst_variables tv c vs =
	TVSet.fold (fun tv' new_str ->
		if tv = tv' then
			let str' = cast_to_store c in
			add_store str' new_str
		else
			add_store_variable tv' new_str
	) vs empty_store

and subst_code_store tv c c_str =
	IntMap.fold (fun inte st new_c_str ->
		let inte = subst_integer tv c inte in
		let st = subst_code_type tv c st in
		add_code_location inte st new_c_str
	) c_str IntMap.empty

and subst_prefix tv c prefix =
	let len = List.length prefix.p_kinds in

	let tv = shift_type_variable 0 len tv in
	let c = shift_constructor 0 len c in

	let cstrts = subst_cstrts tv c prefix.p_cstrts in

	let prefix = { prefix with p_cstrts = cstrts; } in
	(tv, c, prefix)

and subst_cstrts tv c cstrts =
	List.map (subst_cstrt tv c) cstrts

and subst_cstrt tv c cstrt =
	let ics = subst_integer_cstrts tv c cstrt.c_ics in
	let store = subst_store tv c cstrt.c_store in
	let code_store = subst_code_store tv c cstrt.c_code_store in

	let cstrt = { c_ics = ics; c_store = store; c_code_store = code_store; } in
	cstrt

(* [tv -> a]t *)
and subst_type tv c t =
	match t with
	| Name nt ->
		let nt = subst_named_type tv c nt in
		Name nt
	| Concrete ct ->
		let ct = subst_concrete_type tv c ct in
		Concrete ct

and subst_heap_type tv c = function
	| Array at ->
		let at = subst_array_type tv c at in
		Array at
	| Stack st ->
		let st = subst_rich_stack_type tv c st in
		Stack st

and subst_array_type tv c at =
	let t = at.at_type in
	let len = at.at_len in

	let t = subst_type tv c t in
	let len = subst_integer tv c len in

	{ at_type = t; at_len = len; }

and subst_code_type tv c ct =
	match ct with
	| NameCode nt ->
		let nt = subst_named_type tv c nt in
		NameCode nt
	| State st ->
		let st = subst_state_type tv c st in
		State st

and subst_named_type tv c nt =
	let n, cs = nt in
	let cs = List.map (subst_constructor tv c) cs in
	(n, cs)

and subst_constructor tv c = function
	| CStack st ->
		let st = subst_stack_type tv c st in
		CStack st
	| CInt inte ->
		let inte = subst_integer tv c inte in
		CInt inte
	| CStore str ->
		let str = subst_store tv c str in
		CStore str

let subst_named_type_body tv c bdy =
	match bdy with
	| NtdConcrete ct ->
		let ct = subst_concrete_type tv c ct in
		NtdConcrete ct
	| NtdState st ->
		let st = subst_state_type tv c st in
		NtdState st

let subst_named_type_definition tv c ntd =
	let len = List.length ntd.ntd_kinds in

	let tv = shift_type_variable 0 len tv in
	let c = shift_constructor 0 len c in

	let bdy = subst_named_type_body tv c ntd.ntd_body in

	{ ntd with ntd_body = bdy; }


(************************************
     De Bruijn's type application
 ************************************)

exception Too_many_arguments
exception Too_few_arguments
exception Store_mismatch of store

(* (Forall(0, 1, 2, 3 | e).regs)[c] => Forall(0, 1, 2 | e[c/2]).regs[c/2] *)
let apply_any_1 get_kinds set_kinds subst shift a c =
	let kinds = get_kinds a in

	if kinds = [] then raise Too_many_arguments;

	let kinds = Util.remove_last kinds in

	let a = set_kinds kinds a in

	let c = shift_constructor 0 1 c in
	let a = subst (TV 0) c a in
	let a = shift 0 (-1) a in

	a

let apply_state_type_1 = apply_any_1
	(fun state -> state.st_prefix.p_kinds)
	(fun kinds state ->
		let prefix = { state.st_prefix with p_kinds = kinds; } in
		{ state with st_prefix = prefix; })
	subst_state_type
	shift_state_type

let apply_named_type_definition_1 = apply_any_1
	(fun ntd -> ntd.ntd_kinds)
	(fun kinds ntd -> { ntd with ntd_kinds = kinds; })
	subst_named_type_definition
	shift_named_type_definition

let apply_pack_type_1 = apply_any_1
	(fun pt -> pt.pt_prefix.p_kinds)
	(fun kinds pt ->
		let prefix = { pt.pt_prefix with p_kinds = kinds; } in
		{ pt with pt_prefix = prefix; })
	subst_pack_type
	shift_pack_type

let apply_state_type_n cs state =
	List.fold_left apply_state_type_1 state cs

let apply_named_type_definition_n cs ntd =
	List.fold_left apply_named_type_definition_1 ntd cs

let apply_pack_type_n cs pt =
	List.fold_left apply_pack_type_1 pt cs

let apply_any apply_n get_prefix get a cs =
	let a = apply_n cs a in
	let prefix = get_prefix a in

	let kinds = prefix.p_kinds in

	if kinds <> [] then
		raise Too_few_arguments
	else

	let cstrts = prefix.p_cstrts in

	(cstrts, get a)

let apply_named_type_definition ntd cs =
	let (_, nbd) = apply_any
		apply_named_type_definition_n
		(fun ntd -> { p_kinds = ntd.ntd_kinds; p_cstrts = []; })
		(fun ntd -> ntd.ntd_body)
		ntd cs in
	nbd

let apply_pack_type = apply_any
	apply_pack_type_n
	(fun pt -> pt.pt_prefix)
	(fun pt -> pt.pt_tuple)



(************************************
	Size of Types
 ************************************)
open Integer_op

let size_of_tuple_elem_type (sz, _) =
	match sz with
	| Instructions.OpSize8 -> 1
	| Instructions.OpSize16 -> 2
	| Instructions.OpSize32 -> 4
	| Instructions.OpSize48 -> 6
	| Instructions.OpSize64 -> 8
	| Instructions.OpSize128 -> 16

let size_of_tuple_type tt =
	List.fold_left (fun r te ->
		r + (size_of_tuple_elem_type te)
	) 0 tt

let size_of_pack_type pt = 
	size_of_tuple_type pt.pt_tuple

let size_of_concrete_type = function
	| Pack pt ->
		size_of_pack_type pt
	| Tuple tt ->
		size_of_tuple_type tt

let size_of_named_type nt =
	(* XXX : Size of named type is not implemented! *)
	0

let size_of_type = function
	| Name nt ->
		size_of_named_type nt
	| Concrete ct ->
		size_of_concrete_type ct

let size_of_array_type at =
	let { at_type = t; at_len = len; } = at in
	let i = size_of_type t in
	(ie_int i) *.. (ie_inte len)

let rec raw_size_of_stack_type = function
	| StkNull ->
		0
	| StkVar _ ->
		(* Don't care stack variables. They are the same of store variables *)
		0
	| StkCons (_, stk) ->
		4 + raw_size_of_stack_type stk

let size_of_stack_type stk =
	let i = raw_size_of_stack_type stk in
	ie_int i

let size_of_rich_stack_type stk =
	size_of_stack_type stk.rstk_stk

let size_of_heap_type = function
	| Array at ->
		size_of_array_type at
	| Stack st ->
		size_of_rich_stack_type st

(*
 * Convert type checker context to program state
 *)
let context_to_state ctx =
	let prefix = { p_kinds = ctx.ctx_kinds; p_cstrts = [ctx.ctx_cstrt]; } in
	{ st_prefix = prefix; st_registers = ctx.ctx_registers; }

let state_to_contexts state =
	let prefix = state.st_prefix in
	let regs = state.st_registers in
	let kinds = prefix.p_kinds in
	let cstrts = prefix.p_cstrts in
	List.map (fun cstrt ->
		{ ctx_kinds = kinds; ctx_cstrt = cstrt; ctx_registers = regs; }
	) cstrts



(************************************
	Store manipulation (part2)
 ************************************)

exception FIND_LOCATION_SHORTCUT of integer

let find_location ics inte1 locs =
try
	IntMap.iter (fun inte2 _ ->
		if is_equal_integer ics inte1 inte2 then
			raise (FIND_LOCATION_SHORTCUT inte2)
	) locs;
	raise Not_found
with | FIND_LOCATION_SHORTCUT inte ->
	inte

let sub_locations ics locs1 locs2 =
	IntMap.fold (fun i1 ht1 locs1 ->
	try
		let i2 = find_location ics i1 locs1 in
		let ht2 = IntMap.find i2 locs1 in
		if not (is_equal_heap_type ics ht1 ht2) then
			raise Invalid_store_manipulation
		else
			IntMap.remove i2 locs1
	with Not_found ->
		raise Invalid_store_manipulation
	) locs2 locs1

let sub_variables vs1 vs2 =
	TVSet.fold (fun tv vs1 ->
		if TVSet.mem tv vs1 then
			TVSet.remove tv vs1
		else
			raise Invalid_store_manipulation
	) vs2 vs1

let sub_store ics str1 str2 =
	let locs1 = str1.str_locs in
	let vs1 = str1.str_vs in

	let locs2 = str2.str_locs in
	let vs2 = str2.str_vs in

	let locs = sub_locations ics locs1 locs2 in
	let vs = sub_variables vs1 vs2 in

	{ str_locs = locs; str_vs = vs; }

let sub_code_store ics c_str1 c_str2 =
	IntMap.fold (fun i1 ct1 c_str1 ->
	try
		let i2 = find_location ics i1 c_str1 in
		let ct2 = IntMap.find i2 c_str1 in
		if not (is_equal_code_type ics ct1 ct2) then
			raise Invalid_store_manipulation
		else
			IntMap.remove i2 c_str1
	with Not_found ->
		raise Invalid_store_manipulation
	) c_str2 c_str1
