(*
 * Copyright 2005 Toshiyuki Maeda
 *
 * This file is part of TALK.
 *
 * TALK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * TALK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *)

open Instructions
open Annotations
open Pseudo_insts
open Types
open Types_op
open Typing_common
open Typing_error
open Output_types
open Util
open Integer_op

let get_op_size_from_mem_op mem_op =
	mem_op.mop_size

let get_op_size_from_reg = function
        | Reg8 _ ->
		OpSize8
        | Reg16 _ ->
		OpSize16
        | Reg32 _ ->
		OpSize32
        | Regmm _ ->
		OpSize64
        | Regxmm _ ->
		OpSize128
        | Regseg _ ->
		OpSize16

let get_op_size = function
	| P_Op_Mem (mem_op, _) ->
		get_op_size_from_mem_op mem_op
	| P_Op_Imd (_, sz) ->
		sz
	| P_Op_Reg reg ->
		get_op_size_from_reg reg
(*
 *
 * Routines for registers
 *
 *)

let get_reg reg_kind' reg =
	let (reg_kind, inte) = reg in

	if is_equal_register_kind reg_kind reg_kind' then
		inte
	else
		failwithf_type_error "Register kind didn't match."

let get_integer_by_reg32 regs = function
	| EAX -> get_reg RK_E regs.eax
	| ECX -> get_reg RK_E regs.ecx
	| EDX -> get_reg RK_E regs.edx
	| EBX -> get_reg RK_E regs.ebx
	| ESP -> get_reg RK_E regs.esp
	| EBP -> get_reg RK_E regs.ebp
	| ESI -> get_reg RK_E regs.esi
	| EDI -> get_reg RK_E regs.edi

let get_integer_by_reg8 regs = function
	| AL -> get_reg RK_L regs.eax
	| CL -> get_reg RK_L regs.ecx
	| DL -> get_reg RK_L regs.edx
	| BL -> get_reg RK_L regs.ebx
	| AH -> get_reg RK_H regs.eax
	| CH -> get_reg RK_H regs.ecx
	| DH -> get_reg RK_H regs.edx
	| BH -> get_reg RK_H regs.ebx

let get_integer_by_reg regs = function
	| Reg32 reg32 ->
		get_integer_by_reg32 regs reg32
	| Reg8 reg8 ->
		get_integer_by_reg8 regs reg8
	| _ ->
		failwithf_type_error "Unsupported register operand size."

let set_integer_by_reg32 regs inte = function
	| EAX -> { regs with eax = (RK_E, inte); }
	| ECX -> { regs with ecx = (RK_E, inte); }
	| EDX -> { regs with edx = (RK_E, inte); }
	| EBX -> { regs with ebx = (RK_E, inte); }
	| ESP -> { regs with esp = (RK_E, inte); }
	| EBP -> { regs with ebp = (RK_E, inte); }
	| ESI -> { regs with esi = (RK_E, inte); }
	| EDI -> { regs with edi = (RK_E, inte); }

let set_integer_by_reg8 regs inte = function
	| AL -> { regs with eax = (RK_L, inte); }
	| CL -> { regs with ecx = (RK_L, inte); }
	| DL -> { regs with edx = (RK_L, inte); }
	| BL -> { regs with ebx = (RK_L, inte); }
	| AH -> { regs with eax = (RK_H, inte); }
	| CH -> { regs with ecx = (RK_H, inte); }
	| DH -> { regs with edx = (RK_H, inte); }
	| BH -> { regs with ebx = (RK_H, inte); }

let set_integer_by_reg regs inte = function
	| Reg32 reg32 ->
		set_integer_by_reg32 regs inte reg32
	| Reg8 reg8 ->
		set_integer_by_reg8 regs inte reg8
	| _ ->
		failwithf_type_error "Invalid operand size."




(*
 *
 * Routines for modrm and SIB
 *
 *)

let get_disp_opt signed disp_opt =
	match disp_opt with
	| None ->
		None
	| Some v ->
		let converter =
			if signed then Big_int_util.big_int_of_nativeint_signed
			else Big_int_util.big_int_of_nativeint_unsigned
		in
		let v = converter v in
		Some (IEInte (IConst v))

let get_base_opt ctx base_opt =
	match base_opt with
	| None ->
		None
	| Some reg ->
		let ad = get_integer_by_reg ctx.ctx_registers reg in
		let ie = IEInte ad in
		Some ie

let get_scaled_index ctx scaled_index =
	let scale = scaled_index.msi_scale in
	let inte = get_integer_by_reg ctx.ctx_registers scaled_index.msi_index in
	let ie = IEInte inte in

	if scale = 1 then
		ie
	else if (scale = 2) || (scale = 4) || (scale = 8) then
		let scale_inte = IConst (Big_int.big_int_of_int scale) in
		let scale_ie = IEInte scale_inte in
		IEArith (A_MUL, ie, scale_ie)
	else
		failwithf_type_error "Invalid scale %d." scale

let get_scaled_index_opt ctx scaled_index_opt =
	match scaled_index_opt with
	| None ->
		None
	| Some scaled_index ->
		let ad = get_scaled_index ctx scaled_index in
		Some ad



(*
 *
 * Routines for integer (and integer option)
 *
 *)

let rec sum_inte_e_opt ie_opts =
	List.fold_left (fun res ie ->
		match (res, ie) with
		| None, None ->
			None
		| Some ie1, None ->
			Some ie1
		| None, Some ie2 ->
			Some ie2
		| Some ie1, Some ie2 ->
			Some (IEArith (A_ADD, ie1, ie2))
	) None ie_opts

let inte_e_opt_to_inte_e = function
	| None ->
		failwithf_type_error "Invalid mod/rm or SIB."
	| Some ie ->
		ie



(*
 *
 * Routines for memory
 *
 *)

let mem_to_inte_expr ctx mem =
	let base_opt = get_base_opt ctx mem.mop_base_opt in
	let scaled_index_opt = get_scaled_index_opt ctx mem.mop_scaled_index_opt in
	let signed = (base_opt <> None) || (scaled_index_opt <> None) in
	let disp_opt = get_disp_opt signed mem.mop_disp_opt in

	let inte_e_opt = sum_inte_e_opt [base_opt; scaled_index_opt; disp_opt ] in
	inte_e_opt_to_inte_e inte_e_opt
(*
	match mem.mop_size with
	| OpSize32 ->
		let base_opt = get_base_opt ctx mem.mop_base_opt in
		let scaled_index_opt = get_scaled_index_opt ctx mem.mop_scaled_index_opt in
		let signed = (base_opt <> None) || (scaled_index_opt <> None) in
		let disp_opt = get_disp_opt signed mem.mop_disp_opt in

		let inte_e_opt = sum_inte_e_opt [base_opt; scaled_index_opt; disp_opt ] in
		inte_e_opt_to_inte_e inte_e_opt
	| _ ->
		failwithf_type_error "Unsupported memory operand size."
*)

let ad_to_inte_expr ad =
(*
	let scaled_index = (ie_inte (IConst ad.ad_scale)) *.. (ie_inte ad.ad_index) in
	let ie = (ie_inte ad.ad_base) +.. scaled_index in
	ie +.. (ie_inte (IConst ad.ad_offset))
*)
	(ie_inte ad.ad_base) +.. (ie_inte (IConst ad.ad_offset))

let verify_equality ctx mem ad =
	let ie1 = mem_to_inte_expr ctx mem in
	let ie2 = ad_to_inte_expr ad in
	let ic = ie1 =.. ie2 in
	let ics = ctx.ctx_cstrt.c_ics in
	let str = ctx.ctx_cstrt.c_store in

	if not (Constraint_solver.implies_one ics ic) then
		failwithf_type_error "@[Operand (%a) and annotations (%a) didn't match.@]"
			output_inte_expr (ctx, ie1) output_inte_expr (ctx, ie2)

let get_word_type_by_ad ctx ad sz =
	let ics = ctx.ctx_cstrt.c_ics in
	let store = ctx.ctx_cstrt.c_store in
	try
		Store.get ics store ad sz
	with Store_error.Invalid_location_type atyp ->
		failwithf_type_error "@[Load: Invalid Location Type:@ %a@@]" output_at (ctx, atyp)
	| Store_error.Unknown_location inte ->
		failwithf_type_error "@[Load: Unknown Location:@ %a@]" output_inte (ctx, inte)


let get_integer ctx = function
	| P_Op_Mem (mem, ad) ->
		verify_equality ctx mem ad;
		get_word_type_by_ad ctx ad mem.mop_size
	| P_Op_Reg reg ->
		get_integer_by_reg ctx.ctx_registers reg
	| P_Op_Imd (v, _) ->
		IConst v

let set_integer_by_ad ctx inte ad sz =
	let cstrt = ctx.ctx_cstrt in
	let ics = cstrt.c_ics in
	let store = cstrt.c_store in

	try
		let store = Store.set ics store inte ad sz in
		let cstrt = { cstrt with c_store = store; } in
		{ ctx with ctx_cstrt = cstrt; }
	with Store_error.Invalid_location_type atyp ->
		failwithf_type_error "@[Store: Invalid Location Type:@ %a@ where context is@ %a@]"
						output_at (ctx, atyp)
						output_context ctx
	| Store_error.Unknown_location inte ->
		failwithf_type_error "@[Store: Unknown Location:@ %a@]" output_inte (ctx, inte)

let set_integer ctx inte = function
	| P_Op_Mem (mem, ad) ->
		verify_equality ctx mem ad;
		set_integer_by_ad ctx inte ad mem.mop_size
	| P_Op_Reg reg ->
		let regs = set_integer_by_reg ctx.ctx_registers inte reg in
		{ ctx with ctx_registers = regs; }
	| P_Op_Imd _ ->
		failwithf_type_error "Invalid destination operand."


let remove_from_context ctx inte =
	let kinds = ctx.ctx_kinds in
	let cstrt = ctx.ctx_cstrt in
	let store = cstrt.c_store in
	let ics = cstrt.c_ics in

 	(
	match inte with
	| IVar tv ->
		let TV(i) = tv in
		let (kind, _) = List.nth kinds i in
		if kind <> KInt then
			failwithf_type_error "Type variable %a should have a kind KInt."
				Output_types.output_tv (ctx, tv)
	| _ ->
		()
	);

	let typ, store = Store.remove_by_integer ics store inte in

	let cstrt = { cstrt with c_store = store; } in
	let ctx = { ctx with ctx_cstrt = cstrt; } in

	typ, ctx


(*
 * Allocate new TV
 *)

let rec alloc_create_tvs n =
	if n <= 0 then []
	else
	
	TV (n-1) :: alloc_create_tvs (n-1)

let alloc_tvs ctx ks =
	let kinds = ctx.ctx_kinds in
	let regs = ctx.ctx_registers in
	let cstrt = ctx.ctx_cstrt in

	let store = cstrt.c_store in
	let code_store = cstrt.c_code_store in
	let ics = cstrt.c_ics in

	let n = List.length ks in

	let kinds = List.rev_append ks kinds in
	let store = shift_store 0 n store in
	let code_store = shift_code_store 0 n code_store in
	let ics = shift_integer_cstrts 0 n ics in
	let regs = shift_registers_type 0 n regs in

	let cstrt = { c_store = store; c_code_store = code_store; c_ics = ics; } in
	let ctx = { ctx_kinds = kinds; ctx_cstrt = cstrt; ctx_registers = regs; } in

	let tvs = alloc_create_tvs n in
	tvs, ctx

let alloc_inte_tvs dbg_names ctx =
	let ks = List.map (fun dbg_name -> (KInt, dbg_name)) dbg_names in
	let tvs, ctx = alloc_tvs ctx ks in
	let itvs = List.map (fun tv -> (IVar tv)) tvs in
	itvs, ctx

let alloc_inte_tv dbg_name ctx =
	let itvs, ctx = alloc_inte_tvs [dbg_name] ctx in
	(List.nth itvs 0), ctx

let rec get_n_from_dnames dnames n =
	if n <= 0 then []
	else

	let dname, dnames = Util.pop_with_def "__" dnames in
	dname :: (get_n_from_dnames dnames (n - 1))

let alloc_inte_tvs_with_dnames dnames n ctx =
	let dbg_names = get_n_from_dnames dnames n in
	alloc_inte_tvs dbg_names ctx

let alloc_inte_tv_with_dnames dnames ctx =
	let itvs, ctx = alloc_inte_tvs_with_dnames dnames 1 ctx in
	(List.nth itvs 0), ctx

let alloc_inte_two_tvs_with_dnames dnames ctx =
	let itvs, ctx = alloc_inte_tvs_with_dnames dnames 2 ctx in
	let r = (List.nth itvs 0), (List.nth itvs 1) in
	r, ctx
